"""
Utilities for dealing with Chile's RUT ("Rol Único Tributario").

The terms RUT and RUN ("Rol Único Nacional") may be used interchangeably but
only when the holder is a natural person ("persona natural"); a legal person
("persona jurídica") does not have a RUN.

RUT "canonical format": no dots ('.'), with dash ('-'), uppercase K e.g.
``'76042235-5'``, ``'96874030-K'``.

"""

from __future__ import annotations

import itertools
import random
import re

from . import constants


class Rut:
    """
    Representation of a RUT.

    It verifies that the input is syntactically valid and, optionally, that the
    "digito verificador" is correct.

    It does NOT check that the value is within boundaries deemed acceptable by
    the SII (although the regex used does implicitly impose some) nor that the
    RUT has actually been assigned to some person or entity.

    >>> Rut('96874030-K')
    Rut('96874030-K')>
    >>> str(Rut('96874030-K'))
    '96874030-K'
    >>> Rut('96874030-K').digits
    '96874030'
    >>> Rut('96874030-K').dv
    'K'
    >>> Rut('96874030-K').canonical
    '96874030-K'
    >>> Rut('96874030-K').verbose
    '96.874.030-K'
    >>> Rut('96874030-K').digits_with_dots
    '96.874.030'

    >>> Rut('77879240-0') == Rut('77.879.240-0')
    True
    >>> Rut('96874030-K') == Rut('9.68.7403.0-k')
    True

    """

    def __init__(self, value: str | Rut, validate_dv: bool = False) -> None:
        """
        Constructor.

        :param value: a string that represents a syntactically valid RUT
        :param validate_dv: whether to validate that the RUT's
            "digito verificador" is correct

        :raises ValueError:
        :raises TypeError:

        """
        invalid_rut_msg = "Syntactically invalid RUT."

        if isinstance(value, Rut):
            value = value.canonical
        if not isinstance(value, str):
            raise TypeError("Invalid type.")

        clean_value = Rut.clean_str(value)
        match_obj = constants.RUT_CANONICAL_STRICT_REGEX.match(clean_value)
        if match_obj is None:
            raise ValueError(invalid_rut_msg, value)

        match_groups = match_obj.groupdict()
        self._digits = match_groups['digits']
        self._dv = match_groups['dv']

        if validate_dv:
            self.validate_dv(raise_exception=True)

    ############################################################################
    # properties
    ############################################################################

    @property
    def canonical(self) -> str:
        return f'{self._digits}-{self._dv}'

    @property
    def verbose(self) -> str:
        return f'{self.digits_with_dots}-{self._dv}'

    @property
    def digits(self) -> str:
        return self._digits

    @property
    def digits_with_dots(self) -> str:
        """Return RUT digits with a dot ('.') as thousands separator."""
        # > The ',' option signals the use of a comma for a thousands separator.
        #   https://docs.python.org/3/library/string.html#format-specification-mini-language
        return '{:,}'.format(int(self.digits)).replace(',', '.')

    @property
    def dv(self) -> str:
        return self._dv

    ############################################################################
    # magic methods
    ############################################################################

    def __str__(self) -> str:
        return self.canonical

    def __repr__(self) -> str:
        return f"Rut('{self.canonical}')"

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Rut):
            return int(self.digits) < int(other.digits)
        else:
            return NotImplemented

    def __le__(self, other: object) -> bool:
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Rut):
            return self.canonical == other.canonical
        return False

    def __hash__(self) -> int:
        # Objects are hashable so they can be used in hashable collections.
        return hash(self.canonical)

    ############################################################################
    # custom methods
    ############################################################################

    def validate_dv(self, raise_exception: bool = False) -> bool:
        """
        Whether the "digito verificador" of the RUT is correct.

        :param raise_exception: Whether to raise an exception if validation fails.
        :raises ValueError:
        """
        is_valid = self.calc_dv(self._digits) == self._dv
        if not is_valid and raise_exception:
            raise ValueError("RUT's \"digito verificador\" is incorrect.", self.canonical)
        return is_valid

    ############################################################################
    # class methods
    ############################################################################

    @classmethod
    def clean_str(cls, value: str) -> str:
        # note: unfortunately `value.strip('.')` does not remove all the occurrences of '.' in
        #   'value' (only the leading and trailing ones).
        clean_value = value.strip().replace('.', '').upper()
        # Remove leading zeros except if zero is the only digit, so we can accept the RUT '0-0'.
        leading_zero_free_value = re.sub(r'^0+(\d+)', r'\1', clean_value)
        return leading_zero_free_value

    @classmethod
    def calc_dv(cls, rut_digits: str) -> str:
        """
        Calculate the "digito verificador" of a RUT's digits.

        >>> Rut.calc_dv('60910000')
        '1'
        >>> Rut.calc_dv('76555835')
        '2'
        >>> Rut.calc_dv('76177907')
        '9'
        >>> Rut.calc_dv('76369187')
        'K'
        >>> Rut.calc_dv('77879240')
        '0'
        >>> Rut.calc_dv('96874030')
        'K'

        """
        if rut_digits.strip().isdigit() is False:
            raise ValueError("Must be a sequence of digits.")

        # Based on:
        #   https://gist.github.com/rbonvall/464824/4b07668b83ee45121345e4634ebce10dc6412ba3
        s = sum(d * f for d, f in zip(map(int, reversed(rut_digits)), itertools.cycle(range(2, 8))))
        result_alg = 11 - (s % 11)
        return {10: 'K', 11: '0'}.get(result_alg, str(result_alg))

    @classmethod
    def random(cls) -> 'Rut':
        """
        Generate a random RUT.

        Value will be within proper boundaries and "digito verificador"
        will be calculated appropriately i.e. it is not random.

        """
        rut_digits = str(
            random.randint(
                constants.RUT_DIGITS_MIN_VALUE,
                constants.RUT_DIGITS_MAX_VALUE,
            )
        )
        rut_dv = Rut.calc_dv(rut_digits)
        return Rut(f'{rut_digits}-{rut_dv}')



from __future__ import annotations


try:
    import rest_framework
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'djangorestframework' is required to use this module.") from exc
try:
    import django
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'Django' is required to use this module.") from exc

from typing import Mapping, Type

import django.db.models
import rest_framework.serializers

import cl_sii.extras.dj_model_fields
import cl_sii.extras.drf_fields


model_serializer_field_mapping: Mapping[
    Type[django.db.models.Field], Type[rest_framework.serializers.Field]
]
"""
Mapping of Django model fields to DRF serializer fields.

Use this to extend DRF serializers that inherit from :class:`ModelSerializer` so
that Django model fields from :mod:`cl_sii.extras.dj_model_fields` do not have
to be explicitly defined in the serializer.

Usage example:

>>> class ExampleSerializer(rest_framework.serializers.ModelSerializer):
...     serializer_field_mapping = {
...         **rest_framework.serializers.ModelSerializer.serializer_field_mapping,
...         **model_serializer_field_mapping,
...     }
"""
model_serializer_field_mapping = {
    cl_sii.extras.dj_model_fields.RutField: cl_sii.extras.drf_fields.RutField,
}



"""
cl_sii "extras" / Django URL converters.
"""

from __future__ import annotations

from typing import ClassVar

import cl_sii.dte.constants
import cl_sii.rut


class RutConverter:
    """
    Django URL path converter for Chilean RUT.

    Thousands separators are not supported.

    Example:

    >>> from django.urls import path, register_converter
    >>> register_converter(RutConverter, 'cl_sii_rut')
    >>> urlpatterns = [path('example/<cl_sii_rut:emisor_rut>/', ...)]

    .. seealso::
        https://docs.djangoproject.com/en/4.2/topics/http/urls/#registering-custom-path-converters
    """

    regex: ClassVar[str] = r'\d{1,8}-[\dKk]'

    def to_python(self, value: str) -> cl_sii.rut.Rut:
        return cl_sii.rut.Rut(value)

    def to_url(self, value: cl_sii.rut.Rut) -> str:
        return str(value)


class TipoDteConverter:
    """
    Django URL path converter for `Tipo DTE` object.

    Example:

    >>> from django.urls import path, register_converter
    >>> register_converter(TipoDteConverter, 'cl_sii_tipo_dte')
    >>> urlpatterns = [path('example/<cl_sii_tipo_dte:tipo_dte>/', ...)]

    .. seealso::
        https://docs.djangoproject.com/en/4.2/topics/http/urls/#registering-custom-path-converters
    """

    regex: ClassVar[str] = r'\d{2,3}'

    def to_python(self, value: str) -> cl_sii.dte.constants.TipoDte:
        return cl_sii.dte.constants.TipoDte(int(value))

    def to_url(self, value: cl_sii.dte.constants.TipoDte) -> str:
        return str(value.value)



"""
Parse RCV files (CSV)
=====================


"""

import csv
import logging
from datetime import date, datetime
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Tuple

import marshmallow

from cl_sii.base.constants import SII_OFFICIAL_TZ
from cl_sii.extras import mm_fields
from cl_sii.libs import csv_utils, mm_utils, rows_processing, tz_utils
from cl_sii.rut import Rut
from .constants import RcEstadoContable, RcvKind
from .data_models import (
    RcNoIncluirDetalleEntry,
    RcPendienteDetalleEntry,
    RcReclamadoDetalleEntry,
    RcRegistroDetalleEntry,
    RcvDetalleEntry,
    RvDetalleEntry,
)


logger = logging.getLogger(__name__)


RcvCsvFileParserType = Callable[
    [Rut, str, int, Optional[int]],
    Iterable[
        Tuple[
            Optional[RcvDetalleEntry],
            int,
            Dict[str, object],
            Dict[str, object],
        ],
    ],
]


def get_rcv_csv_file_parser(
    rcv_kind: RcvKind,
    estado_contable: Optional[RcEstadoContable],
) -> RcvCsvFileParserType:
    """
    Return a function that parses a CSV file of the given :class:`RcvKind` and
    :class:`RcEstadoContable`.

    :raises ValueError:
    :raises Exception: on unrecoverable errors
    """
    parse_func: RcvCsvFileParserType

    if rcv_kind == RcvKind.COMPRAS:
        if estado_contable is None:
            raise ValueError(
                "'estado_contable' must not be None when 'rcv_kind' is 'COMPRAS'.",
            )
        elif estado_contable == RcEstadoContable.REGISTRO:
            parse_func = parse_rcv_compra_registro_csv_file
        elif estado_contable == RcEstadoContable.NO_INCLUIR:
            parse_func = parse_rcv_compra_no_incluir_csv_file
        elif estado_contable == RcEstadoContable.RECLAMADO:
            parse_func = parse_rcv_compra_reclamado_csv_file
        elif estado_contable == RcEstadoContable.PENDIENTE:
            parse_func = parse_rcv_compra_pendiente_csv_file
        else:
            raise Exception(
                "Programming error. No handler for given 'estado_contable'.",
                estado_contable,
            )
    elif rcv_kind == RcvKind.VENTAS:
        if estado_contable is not None:
            raise ValueError("'estado_contable' must be None when 'rcv_kind' is 'VENTAS'.")
        parse_func = parse_rcv_venta_csv_file
    else:
        raise Exception("Programming error. No handler for given 'rcv_kind'.", rcv_kind)

    return parse_func


def parse_rcv_venta_csv_file(
    rut: Rut,
    input_file_path: str,
    n_rows_offset: int = 0,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RvDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RV ("Registro de Ventas") (CSV file).

    """
    # warning: this looks like it would be executed before the iteration begins but it is not.

    schema_context = dict(
        emisor_rut=rut,
    )
    input_csv_row_schema = RcvVentaCsvRowSchema(context=schema_context)

    expected_input_field_names = (
        'Nro',
        'Tipo Doc',  # 'tipo_docto'
        'Tipo Venta',
        'Rut cliente',  # 'receptor_rut'
        'Razon Social',  # 'receptor_razon_social'
        'Folio',  # 'folio'
        'Fecha Docto',  # 'fecha_emision_date'
        'Fecha Recepcion',  # 'fecha_recepcion_dt'
        'Fecha Acuse Recibo',  # 'fecha_acuse_dt'
        'Fecha Reclamo',  # 'fecha_reclamo_dt'
        'Monto Exento',
        'Monto Neto',
        'Monto IVA',
        'Monto total',  # 'monto_total'
        'IVA Retenido Total',
        'IVA Retenido Parcial',
        'IVA no retenido',
        'IVA propio',
        'IVA Terceros',
        'RUT Emisor Liquid. Factura',
        'Neto Comision Liquid. Factura',
        'Exento Comision Liquid. Factura',
        'IVA Comision Liquid. Factura',
        'IVA fuera de plazo',
        'Tipo Docto. Referencia',
        'Folio Docto. Referencia',
        'Num. Ident. Receptor Extranjero',
        'Nacionalidad Receptor Extranjero',
        'Credito empresa constructora',
        'Impto. Zona Franca (Ley 18211)',
        'Garantia Dep. Envases',
        'Indicador Venta sin Costo',
        'Indicador Servicio Periodico',
        'Monto No facturable',
        'Total Monto Periodo',
        'Venta Pasajes Transporte Nacional',
        'Venta Pasajes Transporte Internacional',
        'Numero Interno',
        'Codigo Sucursal',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Imp.',
        'Valor Otro Imp.',
        'Tasa Otro Imp.',
    )

    fields_to_remove_names = (
        'Nro',
        'Tipo Venta',
        'Monto Exento',
        'Monto Neto',
        'Monto IVA',
        'IVA Retenido Total',
        'IVA Retenido Parcial',
        'IVA no retenido',
        'IVA propio',
        'IVA Terceros',
        'RUT Emisor Liquid. Factura',
        'Neto Comision Liquid. Factura',
        'Exento Comision Liquid. Factura',
        'IVA Comision Liquid. Factura',
        'IVA fuera de plazo',
        'Tipo Docto. Referencia',
        'Folio Docto. Referencia',
        'Num. Ident. Receptor Extranjero',
        'Nacionalidad Receptor Extranjero',
        'Credito empresa constructora',
        'Impto. Zona Franca (Ley 18211)',
        'Garantia Dep. Envases',
        'Indicador Venta sin Costo',
        'Indicador Servicio Periodico',
        'Monto No facturable',
        'Total Monto Periodo',
        'Venta Pasajes Transporte Nacional',
        'Venta Pasajes Transporte Internacional',
        'Numero Interno',
        'Codigo Sucursal',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Imp.',
        'Valor Otro Imp.',
        'Tasa Otro Imp.',
    )

    # note: mypy will complain about returned dataclass type mismatch (and it is right to do so)
    #   but we know from logic which subclass of 'RcvDetalleEntry' will be yielded.
    yield from _parse_rcv_csv_file(  # type: ignore
        input_csv_row_schema,
        expected_input_field_names,
        fields_to_remove_names,
        input_file_path,
        n_rows_offset,
        max_n_rows,
    )


def parse_rcv_compra_registro_csv_file(
    rut: Rut,
    input_file_path: str,
    n_rows_offset: int = 0,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RcRegistroDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RC ("Registro de Compras") / "registro" (CSV file).

    """
    # warning: this looks like it would be executed before the iteration begins but it is not.

    schema_context = dict(
        receptor_rut=rut,
    )
    input_csv_row_schema = RcvCompraRegistroCsvRowSchema(context=schema_context)

    expected_input_field_names = (
        'Nro',
        'Tipo Doc',  # 'tipo_docto'
        'Tipo Compra',
        'RUT Proveedor',  # 'emisor_rut'
        'Razon Social',  # 'emisor_razon_social'
        'Folio',  # 'folio'
        'Fecha Docto',  # 'fecha_emision_date'
        'Fecha Recepcion',  # 'fecha_recepcion_dt'
        'Fecha Acuse',  # 'fecha_acuse_dt'
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Total',  # 'monto_total'
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'Tabacos Puros',
        'Tabacos Cigarrillos',
        'Tabacos Elaborados',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    fields_to_remove_names = (
        'Nro',
        'Tipo Compra',
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'Tabacos Puros',
        'Tabacos Cigarrillos',
        'Tabacos Elaborados',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    # note: mypy will complain about returned dataclass type mismatch (and it is right to do so)
    #   but we know from logic which subclass of 'RcvDetalleEntry' will be yielded.
    yield from _parse_rcv_csv_file(  # type: ignore
        input_csv_row_schema,
        expected_input_field_names,
        fields_to_remove_names,
        input_file_path,
        n_rows_offset,
        max_n_rows,
    )


def parse_rcv_compra_no_incluir_csv_file(
    rut: Rut,
    input_file_path: str,
    n_rows_offset: int = 0,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RcNoIncluirDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RC ("Registro de Compras") / "no incluir" (CSV file).

    """
    # warning: this looks like it would be executed before the iteration begins but it is not.

    schema_context = dict(
        receptor_rut=rut,
    )
    input_csv_row_schema = RcvCompraNoIncluirCsvRowSchema(context=schema_context)

    expected_input_field_names = (
        'Nro',
        'Tipo Doc',  # 'tipo_docto'
        'Tipo Compra',
        'RUT Proveedor',  # 'emisor_rut'
        'Razon Social',  # 'emisor_razon_social'
        'Folio',  # 'folio'
        'Fecha Docto',  # 'fecha_emision_date'
        'Fecha Recepcion',  # 'fecha_recepcion_dt'
        'Fecha Acuse',  # 'fecha_acuse_dt'
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Total',  # 'monto_total'
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    fields_to_remove_names = (
        'Nro',
        'Tipo Compra',
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    # note: mypy will complain about returned dataclass type mismatch (and it is right to do so)
    #   but we know from logic which subclass of 'RcvDetalleEntry' will be yielded.
    yield from _parse_rcv_csv_file(  # type: ignore
        input_csv_row_schema,
        expected_input_field_names,
        fields_to_remove_names,
        input_file_path,
        n_rows_offset,
        max_n_rows,
    )


def parse_rcv_compra_reclamado_csv_file(
    rut: Rut,
    input_file_path: str,
    n_rows_offset: int = 0,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RcReclamadoDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RC ("Registro de Compras") / "reclamado" (CSV file).

    """
    # warning: this looks like it would be executed before the iteration begins but it is not.

    schema_context = dict(
        receptor_rut=rut,
    )
    input_csv_row_schema = RcvCompraReclamadoCsvRowSchema(context=schema_context)

    expected_input_field_names = (
        'Nro',
        'Tipo Doc',  # 'tipo_docto'
        'Tipo Compra',
        'RUT Proveedor',  # 'emisor_rut'
        'Razon Social',  # 'emisor_razon_social'
        'Folio',  # 'folio'
        'Fecha Docto',  # 'fecha_emision_date'
        'Fecha Recepcion',  # 'fecha_recepcion_dt'
        'Fecha Reclamo',  # 'fecha_reclamo_dt'
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Total',  # 'monto_total'
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    fields_to_remove_names = (
        'Nro',
        'Tipo Compra',
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    # note: mypy will complain about returned dataclass type mismatch (and it is right to do so)
    #   but we know from logic which subclass of 'RcvDetalleEntry' will be yielded.
    yield from _parse_rcv_csv_file(  # type: ignore
        input_csv_row_schema,
        expected_input_field_names,
        fields_to_remove_names,
        input_file_path,
        n_rows_offset,
        max_n_rows,
    )


def parse_rcv_compra_pendiente_csv_file(
    rut: Rut,
    input_file_path: str,
    n_rows_offset: int = 0,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RcPendienteDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RC ("Registro de Compras") / "pendiente" (CSV file).

    """
    # warning: this looks like it would be executed before the iteration begins but it is not.

    schema_context = dict(
        receptor_rut=rut,
    )
    input_csv_row_schema = RcvCompraPendienteCsvRowSchema(context=schema_context)

    expected_input_field_names = (
        'Nro',
        'Tipo Doc',  # 'tipo_docto'
        'Tipo Compra',
        'RUT Proveedor',  # 'emisor_rut'
        'Razon Social',  # 'emisor_razon_social'
        'Folio',  # 'folio'
        'Fecha Docto',  # 'fecha_emision_date'
        'Fecha Recepcion',  # 'fecha_recepcion_dt'
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Total',  # 'monto_total'
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    fields_to_remove_names = (
        'Nro',
        'Tipo Compra',
        'Monto Exento',
        'Monto Neto',
        'Monto IVA Recuperable',
        'Monto Iva No Recuperable',
        'Codigo IVA No Rec.',
        'Monto Neto Activo Fijo',
        'IVA Activo Fijo',
        'IVA uso Comun',
        'Impto. Sin Derecho a Credito',
        'IVA No Retenido',
        'NCE o NDE sobre Fact. de Compra',
        'Codigo Otro Impuesto',
        'Valor Otro Impuesto',
        'Tasa Otro Impuesto',
    )

    # note: mypy will complain about returned dataclass type mismatch (and it is right to do so)
    #   but we know from logic which subclass of 'RcvDetalleEntry' will be yielded.
    yield from _parse_rcv_csv_file(  # type: ignore
        input_csv_row_schema,
        expected_input_field_names,
        fields_to_remove_names,
        input_file_path,
        n_rows_offset,
        max_n_rows,
    )


###############################################################################
# schemas
###############################################################################


class _RcvCsvRowSchemaBase(marshmallow.Schema):
    @marshmallow.validates_schema(pass_original=True)
    def validate_schema(self, data: dict, original_data: dict, **kwargs: Any) -> None:
        mm_utils.validate_no_unexpected_input_fields(self, data, original_data)

    # @marshmallow.validates('field_x')
    # def validate_field_x(self, value):
    #     pass

    def to_detalle_entry(self, data: dict) -> RcvDetalleEntry:
        raise NotImplementedError


class RcvVentaCsvRowSchema(_RcvCsvRowSchemaBase):
    FIELD_FECHA_RECEPCION_DT_TZ = SII_OFFICIAL_TZ
    FIELD_FECHA_ACUSE_DT_TZ = SII_OFFICIAL_TZ
    FIELD_FECHA_RECLAMO_DT_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # basic fields
    ###########################################################################

    tipo_docto = mm_fields.RcvTipoDoctoField(
        required=True,
        data_key='Tipo Doc',
    )
    folio = marshmallow.fields.Integer(
        required=True,
        data_key='Folio',
    )
    fecha_emision_date = mm_utils.CustomMarshmallowDateField(
        format='%d/%m/%Y',  # e.g. '22/10/2018'
        required=True,
        data_key='Fecha Docto',
    )
    receptor_rut = mm_fields.RutField(
        required=True,
        data_key='Rut cliente',
    )
    monto_total = marshmallow.fields.Integer(
        required=True,
        data_key='Monto total',
    )
    receptor_razon_social = marshmallow.fields.String(
        required=True,
        data_key='Razon Social',
    )

    ###########################################################################
    # fields whose value is set using data passed in the schema context
    ###########################################################################

    emisor_rut = mm_fields.RutField(
        required=True,
    )

    ###########################################################################
    # extra fields: not included in the returned struct
    ###########################################################################

    fecha_recepcion_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        data_key='Fecha Recepcion',
    )
    fecha_acuse_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=False,
        allow_none=True,
        data_key='Fecha Acuse Recibo',
    )
    fecha_reclamo_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=False,
        allow_none=True,
        data_key='Fecha Reclamo',
    )

    @marshmallow.pre_load
    def preprocess(self, in_data: dict, **kwargs: Any) -> dict:
        # note: required fields checks are run later on automatically thus we may not assume that
        #   values of required fields (`required=True`) exist.

        # Set field value only if it was not in the input data.
        in_data.setdefault('emisor_rut', self.context['emisor_rut'])

        # Fix missing/default values.
        if 'Fecha Acuse Recibo' in in_data:
            if in_data['Fecha Acuse Recibo'] == '':
                in_data['Fecha Acuse Recibo'] = None
        if 'Fecha Reclamo' in in_data:
            if in_data['Fecha Reclamo'] == '':
                in_data['Fecha Reclamo'] = None

        return in_data

    @marshmallow.post_load
    def postprocess(self, data: dict, **kwargs: Any) -> dict:
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13'
        data['fecha_recepcion_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
            dt=data['fecha_recepcion_dt'], tz=self.FIELD_FECHA_RECEPCION_DT_TZ
        )
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13-03:00'
        # >>> data['fecha_recepcion_dt'].astimezone(pytz.UTC).isoformat()
        # '2018-10-23T04:54:13+00:00'

        # note: to express this value in another timezone (but the value does not change), do
        #   `dt_obj.astimezone(pytz.timezone('some timezone'))`

        if 'fecha_acuse_dt' in data and data['fecha_acuse_dt']:
            data['fecha_acuse_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
                dt=data['fecha_acuse_dt'], tz=self.FIELD_FECHA_ACUSE_DT_TZ
            )
        if 'fecha_reclamo_dt' in data and data['fecha_reclamo_dt']:
            data['fecha_reclamo_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
                dt=data['fecha_reclamo_dt'], tz=self.FIELD_FECHA_RECLAMO_DT_TZ
            )

        # Remove leading and trailing whitespace.
        data['receptor_razon_social'] = data['receptor_razon_social'].strip()

        return data

    def to_detalle_entry(self, data: dict) -> RvDetalleEntry:
        try:
            emisor_rut: Rut = data['emisor_rut']
            tipo_docto = data['tipo_docto']
            folio: int = data['folio']
            fecha_emision_date: date = data['fecha_emision_date']
            receptor_rut: Rut = data['receptor_rut']
            monto_total: int = data['monto_total']
            receptor_razon_social: str = data['receptor_razon_social']
            fecha_recepcion_dt: datetime = data['fecha_recepcion_dt']
            fecha_acuse_dt: Optional[datetime] = data['fecha_acuse_dt']
            fecha_reclamo_dt: Optional[datetime] = data['fecha_reclamo_dt']
        except KeyError as exc:
            raise ValueError("Programming error: a referenced field is missing.") from exc

        try:
            detalle_entry = RvDetalleEntry(
                emisor_rut=emisor_rut,
                tipo_docto=tipo_docto,
                folio=folio,
                fecha_emision_date=fecha_emision_date,
                receptor_rut=receptor_rut,
                monto_total=monto_total,
                receptor_razon_social=receptor_razon_social,
                fecha_recepcion_dt=fecha_recepcion_dt,
                fecha_acuse_dt=fecha_acuse_dt,
                fecha_reclamo_dt=fecha_reclamo_dt,
            )
        except (TypeError, ValueError):
            raise

        return detalle_entry


class RcvCompraRegistroCsvRowSchema(_RcvCsvRowSchemaBase):
    FIELD_FECHA_RECEPCION_DT_TZ = SII_OFFICIAL_TZ
    FIELD_FECHA_ACUSE_DT_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # basic fields
    ###########################################################################

    emisor_rut = mm_fields.RutField(
        required=True,
        data_key='RUT Proveedor',
    )
    tipo_docto = mm_fields.RcvTipoDoctoField(
        required=True,
        data_key='Tipo Doc',
    )
    folio = marshmallow.fields.Integer(
        required=True,
        data_key='Folio',
    )
    fecha_emision_date = mm_utils.CustomMarshmallowDateField(
        format='%d/%m/%Y',  # e.g. '22/10/2018'
        required=True,
        data_key='Fecha Docto',
    )
    monto_total = marshmallow.fields.Integer(
        required=True,
        data_key='Monto Total',
    )
    emisor_razon_social = marshmallow.fields.String(
        required=True,
        data_key='Razon Social',
    )

    ###########################################################################
    # fields whose value is set using data passed in the schema context
    ###########################################################################

    receptor_rut = mm_fields.RutField(
        required=True,
    )

    ###########################################################################
    # extra fields: not included in the returned struct
    ###########################################################################

    fecha_recepcion_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        data_key='Fecha Recepcion',
    )
    fecha_acuse_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        allow_none=True,
        data_key='Fecha Acuse',
    )

    @marshmallow.pre_load
    def preprocess(self, in_data: dict, **kwargs: Any) -> dict:
        # note: required fields checks are run later on automatically thus we may not assume that
        #   values of required fields (`required=True`) exist.

        # Set field value only if it was not in the input data.
        in_data.setdefault('receptor_rut', self.context['receptor_rut'])

        # Fix missing/default values.
        if 'Fecha Acuse' in in_data:
            if in_data['Fecha Acuse'] == '':
                in_data['Fecha Acuse'] = None

        return in_data

    @marshmallow.post_load
    def postprocess(self, data: dict, **kwargs: Any) -> dict:
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13'
        data['fecha_recepcion_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
            dt=data['fecha_recepcion_dt'], tz=self.FIELD_FECHA_RECEPCION_DT_TZ
        )
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13-03:00'
        # >>> data['fecha_recepcion_dt'].astimezone(pytz.UTC).isoformat()
        # '2018-10-23T04:54:13+00:00'

        if data['fecha_acuse_dt']:
            data['fecha_acuse_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
                dt=data['fecha_acuse_dt'], tz=self.FIELD_FECHA_ACUSE_DT_TZ
            )

        # note: to express this value in another timezone (but the value does not change), do
        #   `dt_obj.astimezone(pytz.timezone('some timezone'))`

        # Remove leading and trailing whitespace.
        data['emisor_razon_social'] = data['emisor_razon_social'].strip()

        return data

    def to_detalle_entry(self, data: dict) -> RcRegistroDetalleEntry:
        try:
            emisor_rut: Rut = data['emisor_rut']
            tipo_docto = data['tipo_docto']
            folio: int = data['folio']
            fecha_emision_date: date = data['fecha_emision_date']
            receptor_rut: Rut = data['receptor_rut']
            monto_total: int = data['monto_total']
            emisor_razon_social: str = data['emisor_razon_social']
            fecha_recepcion_dt: datetime = data['fecha_recepcion_dt']
            fecha_acuse_dt: Optional[datetime] = data['fecha_acuse_dt']
        except KeyError as exc:
            raise ValueError("Programming error: a referenced field is missing.") from exc

        try:
            detalle_entry = RcRegistroDetalleEntry(
                emisor_rut=emisor_rut,
                tipo_docto=tipo_docto,
                folio=folio,
                fecha_emision_date=fecha_emision_date,
                receptor_rut=receptor_rut,
                monto_total=monto_total,
                emisor_razon_social=emisor_razon_social,
                fecha_recepcion_dt=fecha_recepcion_dt,
                fecha_acuse_dt=fecha_acuse_dt,
            )
        except (TypeError, ValueError):
            raise

        return detalle_entry


class RcvCompraNoIncluirCsvRowSchema(RcvCompraRegistroCsvRowSchema):
    def to_detalle_entry(self, data: dict) -> RcNoIncluirDetalleEntry:
        try:
            emisor_rut: Rut = data['emisor_rut']
            tipo_docto = data['tipo_docto']
            folio: int = data['folio']
            fecha_emision_date: date = data['fecha_emision_date']
            receptor_rut: Rut = data['receptor_rut']
            monto_total: int = data['monto_total']
            emisor_razon_social: str = data['emisor_razon_social']
            fecha_recepcion_dt: datetime = data['fecha_recepcion_dt']
            fecha_acuse_dt: Optional[datetime] = data['fecha_acuse_dt']
        except KeyError as exc:
            raise ValueError("Programming error: a referenced field is missing.") from exc

        try:
            detalle_entry = RcNoIncluirDetalleEntry(
                emisor_rut=emisor_rut,
                tipo_docto=tipo_docto,
                folio=folio,
                fecha_emision_date=fecha_emision_date,
                receptor_rut=receptor_rut,
                monto_total=monto_total,
                emisor_razon_social=emisor_razon_social,
                fecha_recepcion_dt=fecha_recepcion_dt,
                fecha_acuse_dt=fecha_acuse_dt,
            )
        except (TypeError, ValueError):
            raise

        return detalle_entry


class RcvCompraReclamadoCsvRowSchema(_RcvCsvRowSchemaBase):
    FIELD_FECHA_RECEPCION_DT_TZ = SII_OFFICIAL_TZ
    FIELD_FECHA_RECLAMO_DT_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # basic fields
    ###########################################################################

    emisor_rut = mm_fields.RutField(
        required=True,
        data_key='RUT Proveedor',
    )
    tipo_docto = mm_fields.RcvTipoDoctoField(
        required=True,
        data_key='Tipo Doc',
    )
    folio = marshmallow.fields.Integer(
        required=True,
        data_key='Folio',
    )
    fecha_emision_date = mm_utils.CustomMarshmallowDateField(
        format='%d/%m/%Y',  # e.g. '22/10/2018'
        required=True,
        data_key='Fecha Docto',
    )
    monto_total = marshmallow.fields.Integer(
        required=True,
        data_key='Monto Total',
    )
    emisor_razon_social = marshmallow.fields.String(
        required=True,
        data_key='Razon Social',
    )

    ###########################################################################
    # fields whose value is set using data passed in the schema context
    ###########################################################################

    receptor_rut = mm_fields.RutField(
        required=True,
    )

    ###########################################################################
    # extra fields: not included in the returned struct
    ###########################################################################

    fecha_recepcion_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        data_key='Fecha Recepcion',
    )
    fecha_reclamo_dt = marshmallow.fields.DateTime(
        # note: for some reason the rows with 'tipo_docto' equal to
        #   '<RcvTipoDocto.NOTA_CREDITO_ELECTRONICA: 61>' (and maybe others as well) do not
        #   have this field set (always? we do not know).
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=False,
        allow_none=True,
        data_key='Fecha Reclamo',
    )

    @marshmallow.pre_load
    def preprocess(self, in_data: dict, **kwargs: Any) -> dict:
        # note: required fields checks are run later on automatically thus we may not assume that
        #   values of required fields (`required=True`) exist.

        # Set field value only if it was not in the input data.
        in_data.setdefault('receptor_rut', self.context['receptor_rut'])

        # Fix missing/default values.
        # note: for some reason the rows with 'tipo_docto' equal to
        #   '<RcvTipoDocto.NOTA_CREDITO_ELECTRONICA: 61>' (and maybe others as well) do not
        #   have this field set (always? we do not know).
        if 'Fecha Reclamo' in in_data:
            if in_data['Fecha Reclamo'] == '' or 'null' in in_data['Fecha Reclamo']:
                in_data['Fecha Reclamo'] = None

        return in_data

    @marshmallow.post_load
    def postprocess(self, data: dict, **kwargs: Any) -> dict:
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13'
        data['fecha_recepcion_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
            dt=data['fecha_recepcion_dt'], tz=self.FIELD_FECHA_RECEPCION_DT_TZ
        )
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13-03:00'
        # >>> data['fecha_recepcion_dt'].astimezone(pytz.UTC).isoformat()
        # '2018-10-23T04:54:13+00:00'

        if data['fecha_reclamo_dt']:
            data['fecha_reclamo_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
                dt=data['fecha_reclamo_dt'], tz=self.FIELD_FECHA_RECLAMO_DT_TZ
            )

        # note: to express this value in another timezone (but the value does not change), do
        #   `dt_obj.astimezone(pytz.timezone('some timezone'))`

        # Remove leading and trailing whitespace.
        data['emisor_razon_social'] = data['emisor_razon_social'].strip()

        return data

    def to_detalle_entry(self, data: dict) -> RcReclamadoDetalleEntry:
        try:
            emisor_rut: Rut = data['emisor_rut']
            tipo_docto = data['tipo_docto']
            folio: int = data['folio']
            fecha_emision_date: date = data['fecha_emision_date']
            receptor_rut: Rut = data['receptor_rut']
            monto_total: int = data['monto_total']
            emisor_razon_social: str = data['emisor_razon_social']
            fecha_recepcion_dt: datetime = data['fecha_recepcion_dt']
            fecha_reclamo_dt: Optional[datetime] = data['fecha_reclamo_dt']
        except KeyError as exc:
            raise ValueError("Programming error: a referenced field is missing.") from exc

        try:
            detalle_entry = RcReclamadoDetalleEntry(
                emisor_rut=emisor_rut,
                tipo_docto=tipo_docto,
                folio=folio,
                fecha_emision_date=fecha_emision_date,
                receptor_rut=receptor_rut,
                monto_total=monto_total,
                emisor_razon_social=emisor_razon_social,
                fecha_recepcion_dt=fecha_recepcion_dt,
                fecha_reclamo_dt=fecha_reclamo_dt,
            )
        except (TypeError, ValueError):
            raise

        return detalle_entry


class RcvCompraPendienteCsvRowSchema(_RcvCsvRowSchemaBase):
    FIELD_FECHA_RECEPCION_DT_TZ = SII_OFFICIAL_TZ
    FIELD_FECHA_ACUSE_DT_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # basic fields
    ###########################################################################

    emisor_rut = mm_fields.RutField(
        required=True,
        data_key='RUT Proveedor',
    )
    tipo_docto = mm_fields.RcvTipoDoctoField(
        required=True,
        data_key='Tipo Doc',
    )
    folio = marshmallow.fields.Integer(
        required=True,
        data_key='Folio',
    )
    fecha_emision_date = mm_utils.CustomMarshmallowDateField(
        format='%d/%m/%Y',  # e.g. '22/10/2018'
        required=True,
        data_key='Fecha Docto',
    )
    monto_total = marshmallow.fields.Integer(
        required=True,
        data_key='Monto Total',
    )
    emisor_razon_social = marshmallow.fields.String(
        required=True,
        data_key='Razon Social',
    )

    ###########################################################################
    # fields whose value is set using data passed in the schema context
    ###########################################################################

    receptor_rut = mm_fields.RutField(
        required=True,
    )

    ###########################################################################
    # extra fields: not included in the returned struct
    ###########################################################################

    fecha_recepcion_dt = marshmallow.fields.DateTime(
        format='%d/%m/%Y %H:%M:%S',  # e.g. '23/10/2018 01:54:13'
        required=True,
        data_key='Fecha Recepcion',
    )

    @marshmallow.pre_load
    def preprocess(self, in_data: dict, **kwargs: Any) -> dict:
        # note: required fields checks are run later on automatically thus we may not assume that
        #   values of required fields (`required=True`) exist.

        # Set field value only if it was not in the input data.
        in_data.setdefault('receptor_rut', self.context['receptor_rut'])

        # Fix missing/default values.
        if 'Fecha Acuse' in in_data:
            if in_data['Fecha Acuse'] == '':
                in_data['Fecha Acuse'] = None

        return in_data

    @marshmallow.post_load
    def postprocess(self, data: dict, **kwargs: Any) -> dict:
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13'
        data['fecha_recepcion_dt'] = tz_utils.convert_naive_dt_to_tz_aware(
            dt=data['fecha_recepcion_dt'], tz=self.FIELD_FECHA_RECEPCION_DT_TZ
        )
        # >>> data['fecha_recepcion_dt'].isoformat()
        # '2018-10-23T01:54:13-03:00'
        # >>> data['fecha_recepcion_dt'].astimezone(pytz.UTC).isoformat()
        # '2018-10-23T04:54:13+00:00'

        # note: to express this value in another timezone (but the value does not change), do
        #   `dt_obj.astimezone(pytz.timezone('some timezone'))`

        # Remove leading and trailing whitespace.
        data['emisor_razon_social'] = data['emisor_razon_social'].strip()

        return data

    def to_detalle_entry(self, data: dict) -> RcPendienteDetalleEntry:
        try:
            emisor_rut: Rut = data['emisor_rut']
            tipo_docto = data['tipo_docto']
            folio: int = data['folio']
            fecha_emision_date: date = data['fecha_emision_date']
            receptor_rut: Rut = data['receptor_rut']
            monto_total: int = data['monto_total']
            emisor_razon_social: str = data['emisor_razon_social']
            fecha_recepcion_dt: datetime = data['fecha_recepcion_dt']
        except KeyError as exc:
            raise ValueError("Programming error: a referenced field is missing.") from exc

        try:
            detalle_entry = RcPendienteDetalleEntry(
                emisor_rut=emisor_rut,
                tipo_docto=tipo_docto,
                folio=folio,
                fecha_emision_date=fecha_emision_date,
                receptor_rut=receptor_rut,
                monto_total=monto_total,
                emisor_razon_social=emisor_razon_social,
                fecha_recepcion_dt=fecha_recepcion_dt,
            )
        except (TypeError, ValueError):
            raise

        return detalle_entry


###############################################################################
# helpers
###############################################################################


class _RcvCsvDialect(csv.Dialect):
    """
    CSV dialect of RCV CSV files.

    The properties of this dialect were determined with the help of
    :class:`csv.Sniffer`.

    >>> import gzip
    >>> filename = 'SII-download-RCV-file-http-body-response.csv.gz'
    >>> with gzip.open(filename, 'rt', encoding='utf-8') as f:
    ...     dialect = csv.Sniffer().sniff(f.read(50 * 1024))

    """

    delimiter = ';'
    quotechar = '"'
    escapechar = None
    doublequote = False
    skipinitialspace = False
    lineterminator = '\r\n'
    quoting = csv.QUOTE_MINIMAL


def _parse_rcv_csv_file(
    input_csv_row_schema: _RcvCsvRowSchemaBase,
    expected_input_field_names: Sequence[str],
    fields_to_remove_names: Sequence[str],
    input_file_path: str,
    n_rows_offset: int,
    max_n_rows: Optional[int] = None,
) -> Iterable[Tuple[Optional[RcvDetalleEntry], int, Dict[str, object], Dict[str, object]]]:
    """
    Parse entries from an RC or RV (CSV file).

    Common implementation for the different alternatives that depend on the
    kind of RC and RV.

    """
    for field_to_remove_name in fields_to_remove_names:
        if field_to_remove_name not in expected_input_field_names:
            raise Exception(
                "Programming error: field to remove is not one of the expected ones.",
                field_to_remove_name,
            )

    _CSV_ROW_DICT_EXTRA_FIELDS_KEY = '_extra_csv_fields_data'

    fields_to_remove_names += (_CSV_ROW_DICT_EXTRA_FIELDS_KEY,)  # type: ignore

    input_data_enc = 'utf-8'
    # note:
    #   > If csvfile is a file object, it should be opened with newline=''
    #   https://docs.python.org/3/library/csv.html#csv.reader
    with open(input_file_path, mode='rt', encoding=input_data_enc, newline='') as input_f:
        # Create a CSV reader, with auto-detection of header names (first row).
        csv_reader = csv_utils.create_csv_dict_reader(
            input_f,
            csv_dialect=_RcvCsvDialect,
            row_dict_extra_fields_key=_CSV_ROW_DICT_EXTRA_FIELDS_KEY,
            expected_fields_strict=True,
            expected_field_names=expected_input_field_names,
        )

        g = rows_processing.csv_rows_mm_deserialization_iterator(
            csv_reader,
            row_schema=input_csv_row_schema,
            n_rows_offset=n_rows_offset,
            max_n_rows=max_n_rows,
            fields_to_remove_names=fields_to_remove_names,
        )

        for row_ix, row_data, deserialized_row_data, validation_errors in g:
            entry: Optional[RcvDetalleEntry] = None
            row_errors: Dict[str, object] = {}
            conversion_error = None

            if not validation_errors:
                try:
                    entry = input_csv_row_schema.to_detalle_entry(deserialized_row_data)
                except Exception as exc:
                    conversion_error = str(exc)
                    logger.exception(
                        "Deserialized data to data model instance conversion failed "
                        "(probably a programming error)."
                    )

            # Instead of empty dicts, lists, str, etc, we want to have None.
            if validation_errors:
                row_errors['validation'] = validation_errors
            if conversion_error:
                row_errors['other'] = conversion_error

            yield entry, row_ix, row_data, row_errors



"""
Base / constants
================

"""

import enum

import pytz

from cl_sii.libs.tz_utils import PytzTimezone


TZ_CL_SANTIAGO: PytzTimezone = pytz.timezone('America/Santiago')

SII_OFFICIAL_TZ = TZ_CL_SANTIAGO


@enum.unique
class TipoDocumento(enum.IntEnum):
    """
    Enum of "Tipo de Documento".

    They are defined in a document and also an XML schema.

    Source:
      - from official document "FORMATO DOCUMENTOS TRIBUTARIOS ELECTRÓNICOS (2019-11-15)" (retrieved on 2020-01-30)
        (https://github.com/cl-sii-extraoficial/archivos-oficiales/blob/master/src/docs/dte/formato-dtes/2019-11-15-formato-dtes-v2.2.y.pdf)
      - from official schema ``SiiTypes_v10.xsd``, the XML type (enum) ``DocType``. (retrieved on 2021-03-25)
        (https://github.com/cl-sii-extraoficial/archivos-oficiales/blob/master/src/code/rtc/2019-12-12-schema_cesion/schema_cesion/SiiTypes_v10.xsd)

    Notes:
    * It is assumed that ``DocType`` it is the main enumerative for "Tipos de Documentos"
      because even though it has the same elements as the enumerative ``DOCType``, in its
      description it does not contain the adjective "Electrónico".
      In any case, the governing definition should be taken as the one contained in the official
      document and not the XML schema.
    * The elements of the following enums are strictly subgroups of enum ``DocType``:
      * ``DOCType``: "Todos los tipos de Documentos Tributarios Electronicos".
                     (The subgroup of ``DocType`` that are considered "Electrónicos")
      * ``DTEType``: "Tipos de Documentos Tributarios Electronicos"
                    (The subgroup of ``DOCType`` for the choice ``DTEDefType/Documento``)
      * ``LIQType``: "Tipos de Liquidaciones".
                     (The subgroup of ``DOCType`` for the choice ``DTEDefType/Liquidacion``)
      * ``EXPType``: "Tipos de Facturas de Exportacion".
                     (The subgroup of ``DOCType`` for the choice ``DTEDefType/Exportaciones``)
      * ``DTEFacturasType``: "Tipos de Documentos Tributarios Electronicos"
                              (The subset of elements in ``DTEType`` and ``LIQType`` that are
                               "cedibles" or in other words can be included in a "cesión")
    """  # noqa: E501

    FACTURA = 30
    """Factura."""

    FACTURA_NO_AFECTA_O_EXENTA = 32
    """Factura de venta bienes y servicios no afectos o exentos de IVA."""

    FACTURA_ELECTRONICA = 33
    """Factura electrónica de venta."""

    FACTURA_NO_AFECTA_O_EXENTA_ELECTRONICA = 34
    """Factura electrónica de venta, no afecta o exenta de IVA."""
    # aka 'Factura no Afecta o Exenta Electrónica'
    # aka 'Factura Electrónica de Venta de Bienes y Servicios No afectos o Exento de IVA'

    BOLETA = 35
    """Boleta."""

    BOLETA_EXENTA = 38
    """Boleta exenta."""

    BOLETA_ELECTRONICA = 39
    """Boleta electrónica."""

    BOLETA_EXENTA_ELECTRONICA = 41
    """Boleta exenta electrónica."""

    FACTURA_COMPRA = 45
    """Factura de compra."""

    FACTURA_COMPRA_ELECTRONICA = 46
    """Factura electrónica de compra."""
    # aka 'Factura de Compra Electrónica'
    # Name should have been 'Factura Electrónica de Compra'.

    LIQUIDACION = 103
    """Liquidación."""

    LIQUIDACION_FACTURA = 40
    """Liquidación factura."""

    LIQUIDACION_FACTURA_ELECTRONICA = 43
    """Liquidación-Factura Electrónica."""
    # For more info about a "liquidación-factura [electrónica]" see:
    #   - SII / FAQ / "¿Qué son las Liquidaciones-Factura?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_0247.htm
    #   - SII / FAQ / "¿Para qué se utiliza la Liquidación Factura Electrónica?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_3689.htm
    #   - SII / FAQ / "¿Qué documentos tributarios deben emitir las partes involucradas en un
    #     contrato de consignación?"
    #     http://www.sii.cl/preguntas_frecuentes/catastro/001_012_2619.htm
    #   - SII / resoluciones / "Resolución Exenta SII N°108 del 24 de Octubre del 2005.
    #     Establece normas y procedimientos de operación referente a la emisión de
    #     liquidaciones-facturas electrónicas"
    #     http://www.sii.cl/documentos/resoluciones/2005/reso108.htm

    GUIA_DESPACHO = 50
    """Guía de despacho."""

    GUIA_DESPACHO_ELECTRONICA = 52
    """Guía electrónica de despacho."""
    # aka 'Guía de Despacho Electrónica'

    NOTA_DEBITO = 55
    """Nota de débito."""

    NOTA_DEBITO_ELECTRONICA = 56
    """Nota electrónica de débito."""
    # aka 'Nota de Débito Electrónica'

    NOTA_CREDITO = 60
    """Nota de crédito."""

    NOTA_CREDITO_ELECTRONICA = 61
    """Nota electrónica de crédito."""
    # aka 'Nota de Crédito Electrónica'

    FACTURA_EXPORTACION_ELECTRONICA = 110
    """Factura de exportación electrónica."""

    NOTA_DEBITO_EXPORTACION_ELECTRONICA = 111
    """Nota de débito de exportación electrónica."""

    NOTA_CREDITO_EXPORTACION_ELECTRONICA = 112
    """Nota de crédito de exportación electrónica."""

    ORDEN_COMPRA = 801
    """Orden de compra."""

    NOTA_PEDIDO = 802
    """Nota de pedido."""

    CONTRATO = 803
    """Contrato."""

    RESOLUCION = 804
    """Resolución."""

    PROCESO_CHILE_COMPRA = 805
    """Proceso ChileCompra."""

    FICHA_CHILE_COMPRA = 806
    """Ficha ChileCompra."""

    DOCUMENTO_UNICO_SALIDA = 807
    """Documento Único de Salida (DUS)."""

    GUIA_TRANSPORTE_MARITIMO = 808
    """Guía de transporte marítimo."""
    # aka B/L (Conocimiento de embarque).
    # Source:
    #   - INSTRUCCIONES PARA LLENAR LA DECLARACIÓN JURADA F3602
    #     https://www.sii.cl/pagina/iva/inst_exportaciones_3602.htm

    GUIA_TRANSPORTE_AEREO = 809
    """Guía de transporte aéreo (AWB)."""
    # aka 'Air Will Bill'
    # Source:
    #   - INSTRUCCIONES PARA LLENAR LA DECLARACIÓN JURADA F3602
    #     https://www.sii.cl/pagina/iva/inst_exportaciones_3602.htm

    MANIFIESTO_INTERNACIONAL_CARGA = 810
    """Manifiesto Internacional de Carga (MIC)."""
    # Source:
    #   - INSTRUCCIONES PARA LLENAR LA DECLARACIÓN JURADA F3602
    #     https://www.sii.cl/pagina/iva/inst_exportaciones_3602.htm

    CARTA_PORTE = 811
    """Carta de Porte."""

    RESOLUCION_SNA = 812
    """Resolución del SNA donde califica Servicios de Exportación."""

    PASAPORTE = 813
    """Pasaporte."""

    CERTIFICADO_DEPOSITO_BOLSA_PRODUCTO_CHILE = 814
    """Certificado de Depósito Bolsa Prod. Chile."""

    VALE_PRENDA_BOLSA_PRODUCTO_CHILE = 815
    """Vale de Prenda Bolsa Prod. Chile."""

    CODIGO_INSCRIPCION_REGISTRO_ACUERDOS = 820
    """Código de Inscripción en el Registro de Acuerdos con Plazo de Pago Excepcional."""



"""
cl_sii "extras" / Django REST Framework (DRF) fields.

(for serializers)

"""

try:
    import rest_framework
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'djangorestframework' is required to use this module.") from exc

import rest_framework.fields

from cl_sii.rut import Rut


class RutField(rest_framework.fields.CharField):
    """
    DRF field for RUT.

    Data types:
    * native/primitive/internal/deserialized: :class:`cl_sii.rut.Rut`
    * representation/serialized: str, same as for DRF field
      :class:`rest_framework.fields.CharField`

    It verifies only that the input is syntactically valid; it does NOT check
    that the value is within boundaries deemed acceptable by the SII.

    The field performs some input value cleaning when it is an str;
    for example ``' 1.111.111-k \t '`` is allowed and the resulting value
    is ``Rut('1111111-K')``.

    .. seealso::
        :class:`.dj_model_fields.RutField` and :class:`.mm_fields.RutField`

    Implementation partially inspired in
    :class:`rest_framework.fields.UUIDField`.

    """

    default_error_messages = {
        'invalid': "'{value}' is not a syntactically valid RUT.",
    }

    def to_internal_value(self, data: object) -> Rut:
        """
        Deserialize.

        > Restore a primitive datatype into its internal python representation.

        :raises rest_framework.exceptions.ValidationError:
            if the data can't be converted

        """
        if isinstance(data, Rut):
            converted_data = data
        else:
            try:
                if isinstance(data, str):
                    converted_data = Rut(data, validate_dv=False)
                else:
                    self.fail('invalid', value=data)
            except (AttributeError, TypeError, ValueError):
                self.fail('invalid', value=data)

        return converted_data

    def to_representation(self, value: Rut) -> str:
        """
        Serialize.

        > Convert the initial datatype into a primitive, serializable datatype.

        """
        return value.canonical



"""
DTE data models
===============

Concepts
--------

In the domain of a DTE, a:

* "Vendedor": is who sold goods or services to "deudor" in a
  transaction for which the DTE was issued.
  It *usually* corresponds to the DTE's "emisor", but not always.
* "Deudor": is who purchased goods or services from "vendedor" in a
  transaction for which the DTE was issued.
  It *usually* corresponds to the DTE's "receptor", but not always.

"""

from __future__ import annotations

import dataclasses
import logging
from datetime import date, datetime
from typing import Mapping, Optional, Sequence

import pydantic
from typing_extensions import Self

import cl_sii.contribuyente.constants
import cl_sii.rut.constants
from cl_sii.base.constants import SII_OFFICIAL_TZ
from cl_sii.libs import tz_utils
from cl_sii.rut import Rut
from . import constants
from .constants import CodigoReferencia, TipoDte


logger = logging.getLogger(__name__)


def validate_dte_folio(value: int) -> None:
    """
    Validate value for DTE field ``folio``.

    :raises ValueError:
    :raises TypeError:

    """
    if value < constants.DTE_FOLIO_FIELD_MIN_VALUE or value > constants.DTE_FOLIO_FIELD_MAX_VALUE:
        raise ValueError("Value is out of the valid range for 'folio'.")


def validate_dte_monto_total(value: int, tipo_dte: TipoDte) -> None:
    """
    Validate value for DTE field ``monto_total``.

    :raises ValueError:
    :raises TypeError:

    """
    if (
        value < constants.DTE_MONTO_TOTAL_FIELD_MIN_VALUE
        or value > constants.DTE_MONTO_TOTAL_FIELD_MAX_VALUE
    ):
        raise ValueError("Value is out of the valid range for 'monto_total'.")

    if value < 0 and tipo_dte != TipoDte.LIQUIDACION_FACTURA_ELECTRONICA:
        raise ValueError("Value is out of the valid range for 'monto_total'.")


def validate_contribuyente_razon_social(value: str) -> None:
    """
    Validate value for the "razón social" of a "contribuyente".

    :raises ValueError:
    :raises TypeError:

    """
    if len(value) > len(value.strip()):
        raise ValueError("Value must not have leading or trailing whitespace.")

    if len(value) < 1:
        raise ValueError("Value must not be empty.")

    if len(value) > cl_sii.contribuyente.constants.RAZON_SOCIAL_LONG_MAX_LENGTH:
        raise ValueError("Value exceeds max allowed length.")


def validate_clean_str(value: str) -> None:
    if len(value.strip()) != len(value):
        raise ValueError("Value has leading or trailing whitespace characters.", value)


def validate_non_empty_str(value: str) -> None:
    if len(value.strip()) == 0:
        raise ValueError("String value length (stripped) is 0.")


def validate_non_empty_bytes(value: bytes) -> None:
    # warning: do NOT strip a bytes value because "strip" implies an ASCII-encoded text,
    #   which may not be the case.
    # if len(value.strip()) == 0:
    if len(value) == 0:
        raise ValueError("Bytes value length is 0.")


VALIDATION_CONTEXT_TRUST_INPUT: str = 'trust_input'
"""
Key for the validation context to indicate that the input data is trusted.
"""


def is_input_trusted_according_to_validation_context(
    validation_context: Optional[Mapping[str, object]],
) -> bool:
    """
    Return whether the input data is trusted according to the validation context.

    :param validation_context:
        The validation context of a Pydantic model.
        Get it from ``pydantic.ValidationInfo.context``.

    Example for data classes:

    >>> dte_xml_data_instance_kwargs: Mapping[str, object] = dict(
    ...     emisor_rut=Rut('60910000-1'),  # ...
    ... )
    >>> dte_xml_data_adapter = pydantic.TypeAdapter(DteXmlData)
    >>> dte_xml_data_instance: DteXmlData = dte_xml_data_adapter.validate_python(
    ...     dte_xml_data_instance_kwargs,
    ...     context={VALIDATION_CONTEXT_TRUST_INPUT: True}
    ... )
    """
    if validation_context is None:
        return False
    else:
        return validation_context.get(VALIDATION_CONTEXT_TRUST_INPUT) is True


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteNaturalKey:
    """
    Natural key of a DTE.

    The class instances are immutable.

    This group of fields uniquely identifies a DTE.

    >>> instance = DteNaturalKey(Rut('60910000-1'), TipoDte.FACTURA_ELECTRONICA, 2093465)

    >>> str(instance)
    "DteNaturalKey(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDte.FACTURA_ELECTRONICA: 33>, folio=2093465)"
    >>> str(instance) == repr(instance)
    True
    >>> instance.slug
    '60910000-1--33--2093465'

    """

    emisor_rut: Rut
    """
    RUT of the "emisor" of the DTE.
    """

    tipo_dte: TipoDte
    """
    The kind of DTE.
    """

    folio: int
    """
    The sequential number of a DTE of given kind issued by 'emisor_rut'.
    """

    def as_dict(self) -> Mapping[str, object]:
        return dataclasses.asdict(self)

    @property
    def slug(self) -> str:
        """
        Return an slug representation (that preserves uniquess) of the instance.
        """
        # note: many alternatives were considered and discarded such as:
        #   f'{self.emisor_rut}-{self.tipo_dte}-{self.folio}'
        #   f'{self.emisor_rut}.{self.tipo_dte}.{self.folio}'
        #   f'{self.emisor_rut}/{self.tipo_dte}/{self.folio}'
        #   f'R-{self.emisor_rut}-T-{self.tipo_dte}-F-{self.folio}'
        #   f'rut-{self.emisor_rut}-tipo-{self.tipo_dte}-folio-{self.folio}'

        return f'{self.emisor_rut}--{self.tipo_dte}--{self.folio}'

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('folio')
    @classmethod
    def validate_folio(cls, v: object) -> object:
        if isinstance(v, int):
            validate_dte_folio(v)
        return v


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteDataL0(DteNaturalKey):
    """
    DTE data level 0.

    Its fields are enough to uniquely identify a DTE but nothing more.

    The class instances are immutable.

    >>> instance = DteDataL0(
    ...     Rut('60910000-1'), TipoDte.FACTURA_ELECTRONICA, 2093465, date(2018, 5, 7),
    ...     Rut('60910000-1'), 10403)

    >>> str(instance)
    "DteDataL0(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDte.FACTURA_ELECTRONICA: 33>, " \
    "folio=2093465)"
    >>> str(instance) == repr(instance)
    True
    >>> instance.slug
    '60910000-1--33--2093465'
    >>> instance.natural_key
    "DteNaturalKey(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDte.FACTURA_ELECTRONICA: 33>, folio=2093465)"

    """

    @property
    def natural_key(self) -> DteNaturalKey:
        return DteNaturalKey(emisor_rut=self.emisor_rut, tipo_dte=self.tipo_dte, folio=self.folio)


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteDataL1(DteDataL0):
    """
    DTE data level 1.

    It is the minimal set of DTE data fields that are useful.
    For example, SII has an endpoint that confirms that a given DTE exists,
    and the data that it requires can be obtained from this struct.

    The class instances are immutable.

    >>> instance = DteDataL1(
    ...     Rut('60910000-1'), TipoDte.FACTURA_ELECTRONICA, 2093465, date(2018, 5, 7),
    ...     Rut('60910000-1'), 10403)

    >>> str(instance)
    "DteDataL1(" \
    "emisor_rut=Rut('60910000-1'), tipo_dte=<TipoDte.FACTURA_ELECTRONICA: 33>, " \
    "folio=2093465, fecha_emision_date=datetime.date(2018, 5, 7), " \
    "receptor_rut=Rut('60910000-1'), monto_total=10403)"
    >>> str(instance) == repr(instance)
    True

    """

    fecha_emision_date: date
    """
    Field 'fecha_emision' of the DTE.

    .. warning:: It may not match the **real date** on which the DTE was issued
        or received/processed by SII.

    """

    receptor_rut: Rut
    """
    RUT of the "receptor" of the DTE.
    """

    monto_total: int
    """
    Total amount of the DTE.
    """

    ###########################################################################
    # properties
    ###########################################################################

    @property
    def vendedor_rut(self) -> Rut:
        """
        Return the RUT of the "vendedor" aka "proveedor" (supplier).

        :raises ValueError:
        """
        if self.tipo_dte.emisor_is_vendedor:
            result = self.emisor_rut
        elif self.tipo_dte.receptor_is_vendedor:
            result = self.receptor_rut
        else:
            raise ValueError(
                "Concept \"vendedor\" does not apply for this 'tipo_dte'.", self.tipo_dte
            )

        return result

    @property
    def comprador_rut(self) -> Rut:
        """
        Return the RUT of the "comprador" aka "cliente" (buyer).

        :raises ValueError:
        """
        if self.tipo_dte.emisor_is_vendedor:
            result = self.receptor_rut
        elif self.tipo_dte.receptor_is_vendedor:
            result = self.emisor_rut
        else:
            raise ValueError(
                "Concepts \"comprador\" and \"deudor\" do not apply for this 'tipo_dte'.",
                self.tipo_dte,
            )

        return result

    @property
    def deudor_rut(self) -> Rut:
        """
        Return the RUT of the "deudor" (same as the "comprador").

        :raises ValueError:
        """
        return self.comprador_rut

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('monto_total')
    @classmethod
    def validate_monto_total(cls, v: object, info: pydantic.ValidationInfo) -> object:
        tipo_dte = info.data['tipo_dte']

        if isinstance(v, int) and isinstance(tipo_dte, TipoDte):
            validate_dte_monto_total(v, tipo_dte=tipo_dte)

        return v


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteDataL2(DteDataL1):
    """
    DTE data level 2.

    Very similar to :class:`DteXmlData` (and a lot of duplicated code,
    unfortunately).

    About fields
    - ``emisor_razon_social``: redundant but required by the DTE XML schema.
    - ``receptor_razon_social``: redundant but required by the DTE XML schema.
    - ``fecha_vencimiento`` (date): important for some business logic
      but it is not required by the DTE XML schema.

    The class instances are immutable.

    """

    ###########################################################################
    # constants
    ###########################################################################

    DATETIME_FIELDS_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # fields
    ###########################################################################

    emisor_razon_social: Optional[str]
    """
    "Razón social" (legal name) of the "emisor" of the DTE.
    """

    receptor_razon_social: Optional[str]
    """
    "Razón social" (legal name) of the "receptor" of the DTE.
    """

    fecha_vencimiento_date: Optional[date] = None
    """
    "Fecha de vencimiento (pago)" of the DTE.
    """

    firma_documento_dt: Optional[datetime] = None
    """
    Datetime on which the "documento" was digitally signed.
    """

    signature_value: Optional[bytes] = None
    """
    DTE's digital signature's value (raw bytes, without base64 encoding).
    """

    signature_x509_cert_der: Optional[bytes] = None
    """
    DTE's digital signature's DER-encoded X.509 cert.

    .. seealso::
        Functions :func:`cl_sii.libs.crypto_utils.load_der_x509_cert`
        and :func:`cl_sii.libs.crypto_utils.x509_cert_der_to_pem`.
    """

    emisor_giro: Optional[str] = None
    """
    "Giro" of the "emisor" of the DTE.
    """

    emisor_email: Optional[str] = None
    """
    Email address of the "emisor" of the DTE.
    """

    receptor_email: Optional[str] = None
    """
    Email address of the "receptor" of the DTE.
    """

    def as_dte_data_l1(self) -> DteDataL1:
        return DteDataL1(
            emisor_rut=self.emisor_rut,
            tipo_dte=self.tipo_dte,
            folio=self.folio,
            fecha_emision_date=self.fecha_emision_date,
            receptor_rut=self.receptor_rut,
            monto_total=self.monto_total,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('emisor_razon_social', 'receptor_razon_social')
    @classmethod
    def validate_contribuyente_razon_social(cls, v: object) -> object:
        if isinstance(v, str):
            validate_contribuyente_razon_social(v)
        return v

    @pydantic.field_validator('firma_documento_dt')
    @classmethod
    def validate_datetime_tz(cls, v: object) -> object:
        if isinstance(v, datetime):
            tz_utils.validate_dt_tz(v, cls.DATETIME_FIELDS_TZ)
        return v

    @pydantic.field_validator('signature_value', 'signature_x509_cert_der')
    @classmethod
    def validate_non_empty_bytes(cls, v: object) -> object:
        if isinstance(v, bytes):
            validate_non_empty_bytes(v)
        return v

    @pydantic.field_validator('emisor_giro', 'emisor_email', 'receptor_email')
    @classmethod
    def validate_no_leading_or_trailing_whitespace_characters(cls, v: object) -> object:
        if isinstance(v, str):
            validate_clean_str(v)
        return v

    @pydantic.field_validator('emisor_giro', 'emisor_email', 'receptor_email')
    @classmethod
    def validate_non_empty_stripped_str(cls, v: object) -> object:
        if isinstance(v, str):
            validate_non_empty_str(v)
        return v


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteXmlReferencia:
    """
    Data in XML element ``Referencia`` in an DTE XML doc.

    > Identificacion de otros documentos Referenciados por Documento

    DTE doc XML element: 'Documento//Referencia'

    .. note:: An XML DTE document includes none or up to 40 "Referencia" elements.

    .. seealso::
        XML schema of ``{http://www.sii.cl/SiiDte}/DTE/Documento/Referencia`` in 'DTE_v10.xsd' at
        https://github.com/cl-sii-extraoficial/archivos-oficiales/blob/master/src/code/dte/README.md

    """

    ###########################################################################
    # Fields
    ###########################################################################

    numero_linea_ref: int
    """
    Sequential line  number of the "referencia". Must be an integer between 1 and 40 inclusive

    > Numero Secuencial de Linea de Referencia

    DTE doc XML element: '..//Documento//Referencia//NroLinRef'
    """

    tipo_documento_ref: str
    """
    Kind of the document of "Referencia". Length must be >= 1 and <= 3

    > Tipo de Documento de Referencia

    DTE doc XML element: '..//Documento//Referencia//TpoDocRef'

    .. note::
        This field accepts any of the elements of the class:`TipoDocumento` or
        an alphanumeric to refer to non-tax documents (in this case, validation does not apply)
    """

    folio_ref: str
    """
    The folio of the document referred to.

    > Identificación del documento de referencia.

    DTE doc XML element: '..//Documento//Referencia//FolioRef'
    """

    fecha_ref: date
    """
    The 'fecha_emision' of the document referred to.

    > Fecha del documento de referencia

    DTE doc XML element: '..//Documento//Referencia//FchRef'
    """

    ind_global: Optional[int] = None
    """
    Whether a set of documents of the same kind is referenced.

    > Documento afecta a un número de más de 20 documentos del mismo `tipo_documento_ref`
    > Se explicita la razón en `Razón Referencia`

    DTE doc XML element: '..//Documento//Referencia//IndGlobal'
    """

    rut_otro: Optional[Rut] = None
    """
    The RUT of the "emisor" of the document referred to.

    > RUT otro contribuyente

    DTE doc XML element: '..//Documento//Referencia//RUTOtr'

    .. note::
        > Sólo si el documento de referencia es de tipo tributario y fue emitido
        > por otro contribuyente
    """

    codigo_ref: Optional[CodigoReferencia] = None
    """
    The type of use for the reference

    > Tipo de Uso de la Referencia

    DTE doc XML element: '..//Documento//Referencia//CodRef'
    """

    razon_ref: Optional[str] = None
    """
    The reason the document is being referenced

    > Razon Explicita por la que se Referencia el Documento

    DTE doc XML element: '..//Documento//Referencia//RazonRef'
    """

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('numero_linea_ref')
    @classmethod
    def validate_numero_linea_ref(cls, value: int) -> int:
        if (
            constants.DTE_REFERENCIA_LINE_NUMBER_MIN_VALUE
            <= value
            <= constants.DTE_REFERENCIA_LINE_NUMBER_MAX_VALUE
        ):
            return value

        raise ValueError(
            "Value 'numero_linea_ref' must be a value between "
            f"{constants.DTE_REFERENCIA_LINE_NUMBER_MIN_VALUE} and "
            f"{constants.DTE_REFERENCIA_LINE_NUMBER_MAX_VALUE}",
            value,
        )

    @pydantic.field_validator('tipo_documento_ref')
    @classmethod
    def validate_tipo_documento_ref(cls, value: str) -> str:
        if 1 <= len(value) <= 3:
            return value

        raise ValueError(
            "The length of 'tipo_documento_ref' must be a value between 1 and 3", value
        )

    @pydantic.field_validator('ind_global')
    @classmethod
    def validate_ind_global(cls, value: int | None) -> int | None:
        if value and value != 1:
            raise ValueError("Only the value '1' is valid for the field 'ind_global'", value)
        return value

    @pydantic.field_validator('folio_ref')
    @classmethod
    def validate_folio_ref(cls, value: str) -> str:
        if (
            constants.DTE_REFERENCIA_FOLIO_MIN_LENGTH
            <= len(value)
            <= constants.DTE_REFERENCIA_FOLIO_MAX_LENGTH
        ):
            return value

        raise ValueError(
            "The length of 'folio_ref' must be a value between "
            f"{constants.DTE_REFERENCIA_FOLIO_MIN_LENGTH} and "
            f"{constants.DTE_REFERENCIA_FOLIO_MAX_LENGTH}",
            value,
        )

    @pydantic.field_validator('fecha_ref')
    @classmethod
    def validate_fecha_ref(cls, value: date) -> date:
        if (
            value < constants.DTE_REFERENCIA_FECHA_NOT_BEFORE
            or value > constants.DTE_REFERENCIA_FECHA_NOT_AFTER
        ):
            raise ValueError(
                "The date 'fecha_ref' must be after "
                f"{constants.DTE_REFERENCIA_FECHA_NOT_BEFORE} and before "
                f"{constants.DTE_REFERENCIA_FECHA_NOT_AFTER}",
                value,
            )

        return value

    @pydantic.field_validator('razon_ref')
    @classmethod
    def validate_razon_ref(cls, value: str | None) -> str | None:
        if value and len(value) > constants.DTE_REFERENCIA_RAZON_MAX_LENGTH:
            raise ValueError(
                "The maximum length allowed for `razon_ref` is "
                f"{constants.DTE_REFERENCIA_RAZON_MAX_LENGTH}",
                value,
            )

        return value


@pydantic.dataclasses.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    ),
)
class DteXmlData(DteDataL1):
    """
    DTE XML data.

    Very similar to :class:`DteDataL2` (and a lot of duplicated code,
    unfortunately).

    About fields
    - ``emisor_razon_social``: redundant but required by the DTE XML schema.
    - ``receptor_razon_social``: redundant but required by the DTE XML schema.
    - ``fecha_vencimiento`` (date): important for some business logic
      but it is not required by the DTE XML schema.

    The class instances are immutable.

    """

    ###########################################################################
    # constants
    ###########################################################################

    DATETIME_FIELDS_TZ = SII_OFFICIAL_TZ

    ###########################################################################
    # fields
    ###########################################################################

    emisor_razon_social: str
    """
    "Razón social" (legal name) of the "emisor" of the DTE.
    """

    receptor_razon_social: str
    """
    "Razón social" (legal name) of the "receptor" of the DTE.
    """

    fecha_vencimiento_date: Optional[date] = None
    """
    "Fecha de vencimiento (pago)" of the DTE.
    """

    firma_documento_dt: Optional[datetime] = None
    """
    Datetime on which the "documento" was digitally signed.
    """

    signature_value: Optional[bytes] = None
    """
    DTE's digital signature's value (raw bytes, without base64 encoding).
    """

    signature_x509_cert_der: Optional[bytes] = None
    """
    DTE's digital signature's DER-encoded X.509 cert.

    .. seealso::
        Functions :func:`cl_sii.libs.crypto_utils.load_der_x509_cert`
        and :func:`cl_sii.libs.crypto_utils.x509_cert_der_to_pem`.
    """

    emisor_giro: Optional[str] = None
    """
    "Giro" of the "emisor" of the DTE.
    """

    emisor_email: Optional[str] = None
    """
    Email address of the "emisor" of the DTE.
    """

    receptor_email: Optional[str] = None
    """
    Email address of the "receptor" of the DTE.
    """

    referencias: Optional[Sequence[DteXmlReferencia]] = None
    """
    List of structs for ``Referencia`` XML elements.

    ..warning::
        The items MUST be ordered according to their ``numero_linea_ref``.
    """

    def as_dte_data_l1(self) -> DteDataL1:
        return DteDataL1(
            emisor_rut=self.emisor_rut,
            tipo_dte=self.tipo_dte,
            folio=self.folio,
            fecha_emision_date=self.fecha_emision_date,
            receptor_rut=self.receptor_rut,
            monto_total=self.monto_total,
        )

    def as_dte_data_l2(self) -> DteDataL2:
        return DteDataL2(
            emisor_rut=self.emisor_rut,
            tipo_dte=self.tipo_dte,
            folio=self.folio,
            fecha_emision_date=self.fecha_emision_date,
            receptor_rut=self.receptor_rut,
            monto_total=self.monto_total,
            emisor_razon_social=self.emisor_razon_social,
            receptor_razon_social=self.receptor_razon_social,
            fecha_vencimiento_date=self.fecha_vencimiento_date,
            firma_documento_dt=self.firma_documento_dt,
            signature_value=self.signature_value,
            signature_x509_cert_der=self.signature_x509_cert_der,
            emisor_giro=self.emisor_giro,
            emisor_email=self.emisor_email,
            receptor_email=self.receptor_email,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('emisor_razon_social', 'receptor_razon_social')
    @classmethod
    def validate_contribuyente_razon_social(cls, v: object) -> object:
        if isinstance(v, str):
            validate_contribuyente_razon_social(v)
        return v

    @pydantic.field_validator('firma_documento_dt')
    @classmethod
    def validate_datetime_tz(cls, v: object) -> object:
        if isinstance(v, datetime):
            tz_utils.validate_dt_tz(v, cls.DATETIME_FIELDS_TZ)
        return v

    @pydantic.field_validator('signature_value', 'signature_x509_cert_der')
    @classmethod
    def validate_non_empty_bytes(cls, v: object) -> object:
        if isinstance(v, bytes):
            validate_non_empty_bytes(v)
        return v

    @pydantic.field_validator('emisor_giro', 'emisor_email', 'receptor_email')
    @classmethod
    def validate_no_leading_or_trailing_whitespace_characters(cls, v: object) -> object:
        if isinstance(v, str):
            validate_clean_str(v)
        return v

    @pydantic.field_validator('emisor_giro', 'emisor_email', 'receptor_email')
    @classmethod
    def validate_non_empty_stripped_str(cls, v: object) -> object:
        if isinstance(v, str):
            validate_non_empty_str(v)
        return v

    @pydantic.field_validator('referencias')
    @classmethod
    def validate_referencias_numero_linea_ref_order(cls, v: object) -> object:
        if isinstance(v, Sequence):
            numero_linea_refs = [referencia.numero_linea_ref for referencia in v]
            if numero_linea_refs != sorted(numero_linea_refs):
                raise ValueError(
                    "items must be ordered according to their 'numero_linea_ref'. "
                    f"All numero_linea_refs: "
                    f"{', '.join(str(num_linea_ref) for num_linea_ref in numero_linea_refs)}"
                )
        return v

    @pydantic.model_validator(mode='after')
    def validate_referencias_rut_otro_is_consistent_with_tipo_dte(
        self, info: pydantic.ValidationInfo
    ) -> Self:
        referencias = self.referencias
        tipo_dte = self.tipo_dte

        if (
            isinstance(referencias, Sequence)
            and isinstance(tipo_dte, TipoDte)
            and tipo_dte not in constants.DTE_REFERENCIA_RUTOTR_TIPO_DOC_SET
        ):
            for referencia in referencias:
                if referencia.rut_otro:
                    message: str = (
                        f"Setting a 'rut_otro' is not a valid option for this 'tipo_dte':"
                        f" 'tipo_dte' == {tipo_dte!r},"
                        f" 'Referencia' number {referencia.numero_linea_ref}."
                    )
                    if is_input_trusted_according_to_validation_context(info.context):
                        logger.warning('Validation failed but input is trusted: %s', message)
                    else:
                        raise ValueError(message)

        return self

    @pydantic.model_validator(mode='after')
    def validate_referencias_rut_otro_is_consistent_with_emisor_rut(
        self, info: pydantic.ValidationInfo
    ) -> Self:
        referencias = self.referencias
        emisor_rut = self.emisor_rut

        if isinstance(referencias, Sequence) and isinstance(emisor_rut, Rut):
            for referencia in referencias:
                if referencia.rut_otro and referencia.rut_otro == emisor_rut:
                    message: str = (
                        f"'rut_otro' must be different from 'emisor_rut':"
                        f" {referencia.rut_otro!r} == {emisor_rut!r},"
                        f" 'Referencia' number {referencia.numero_linea_ref}."
                    )
                    if is_input_trusted_according_to_validation_context(info.context):
                        logger.warning('Validation failed but input is trusted: %s', message)
                    else:
                        raise ValueError(message)

        return self

    @pydantic.model_validator(mode='after')
    def validate_referencias_codigo_ref_is_consistent_with_tipo_dte(self) -> Self:
        referencias = self.referencias
        tipo_dte = self.tipo_dte

        if (
            isinstance(referencias, Sequence)
            and isinstance(tipo_dte, TipoDte)
            and tipo_dte in constants.DTE_REFERENCIA_CODREF_TIPO_DOC_MANDATORY_SET
        ):
            for referencia in referencias:
                if not referencia.codigo_ref:
                    raise ValueError(
                        f"'codigo_ref' is mandatory for this 'tipo_dte':"
                        f" 'tipo_dte' == {tipo_dte!r},"
                        f" 'Referencia' number {referencia.numero_linea_ref}.",
                    )

        return self


DTE_XML_DATA_PYDANTIC_TYPE_ADAPTER = pydantic.TypeAdapter(DteXmlData)



"""
cl_sii "extras" / Django model fields.

"""

try:
    import django
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'Django' is required to use this module.") from exc

from typing import Any, Optional, Tuple

import django.core.exceptions
import django.db.models
import django.db.models.fields

import cl_sii.rut.constants
from cl_sii.rut import Rut


class RutField(django.db.models.Field):
    """
    Django model field for RUT.

    * Python data type: :class:`cl_sii.rut.Rut`
    * DB type: ``varchar``, the same one as the one for model field
      :class:`django.db.models.CharField`

    It verifies only that the input is syntactically valid; it does NOT check
    that the value is within boundaries deemed acceptable by the SII.

    The field performs some input value cleaning when it is an str;
    for example ``' 1.111.111-k \t '`` is allowed and the resulting value
    is ``Rut('1111111-K')``.

    .. seealso::
        :class:`.drf_fields.RutField` and :class:`.mm_fields.RutField`

    Implementation partially inspired in
    :class:`django.db.models.fields.UUIDField`.

    """

    # TODO: add option to validate that "digito verificador" is correct.
    # TODO: implement method 'formfield'. Probably a copy of 'CharField.formfield' is fine.

    description = 'RUT for SII (Chile)'
    default_error_messages = {
        'invalid': "'%(value)s' is not a syntactically valid RUT.",
        'invalid_dv': "\"digito verificador\" of RUT '%(value)s' is incorrect.",
    }
    empty_strings_allowed = False

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # note: the value saved to the DB will always be in canonical format.
        db_column_max_length = cl_sii.rut.constants.RUT_CANONICAL_MAX_LENGTH

        # note: for some reason, even though the kwarg 'max_length' was not set explicitly in
        #   a model, some Django magic caused it was set automatically (perhaps consecutive calls
        #   or something like that?).
        if 'max_length' in kwargs and kwargs['max_length'] != db_column_max_length:
            raise ValueError("This field does not allow customization of 'max_length'.")

        kwargs['max_length'] = db_column_max_length
        super().__init__(*args, **kwargs)

    def deconstruct(self) -> Tuple[str, str, Any, Any]:
        """
        Return a 4-tuple with enough information to recreate the field.
        """
        # note: this override is necessary because we have a custom constructor.

        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']

        return name, path, args, kwargs

    def get_internal_type(self) -> str:
        # Emulate built-in model field type 'CharField' i.e. the underlying DB type is the same.
        #   https://docs.djangoproject.com/en/2.1/howto/custom-model-fields/#emulating-built-in-field-types
        return 'CharField'

    def from_db_value(
        self,
        value: Optional[str],
        expression: object,
        connection: object,
    ) -> Optional[Rut]:
        """
        Convert a value as returned by the database to a Python object.

        > It is the reverse of :meth:`get_prep_value`.

        > If present for the field subclass, :meth:`from_db_value` will be
        > called in all circumstances when the data is loaded from the
        > database, including in aggregates and ``values()`` calls.

        It needs to be able to process ``None``.

        .. seealso::
            https://docs.djangoproject.com/en/2.1/howto/custom-model-fields/#converting-values-to-python-objects
            https://docs.djangoproject.com/en/2.1/ref/models/fields/#django.db.models.Field.from_db_value

        """
        # note: there is no parent implementation, for performance reasons.
        return self.to_python(value)

    def get_prep_value(self, value: Optional[object]) -> Optional[str]:
        """
        Convert the model's attribute value to a format suitable for the DB.

        i.e. prepared for use as a parameter in a query.
        It is the reverse of :meth:`from_db_value`.

        However, these are preliminary non-DB specific value checks and
        conversions (otherwise customize :meth:`get_db_prep_value`).

        Note: Before returning, ``value`` will be passed to :meth:`to_python` so that, if needed, it
        will be converted to an instance of :class:`Rut`, which is very convenient in cases such
        as when the type of ``value`` is :class:`str`.
        """
        value = super().get_prep_value(value)
        value_rut: Optional[Rut] = self.to_python(value)
        return value_rut if value_rut is None else value_rut.canonical

    def to_python(self, value: Optional[object]) -> Optional[Rut]:
        """
        Convert the input value to the correct Python object (:class:`Rut`).

        > It acts as the reverse of :meth:`value_to_string`, and is also
        called in :meth`clean`.

        It needs to be able to process ``None``.

        .. seealso::
            https://docs.djangoproject.com/en/2.1/howto/custom-model-fields/#converting-values-to-python-objects
            https://docs.djangoproject.com/en/2.1/ref/models/fields/#django.db.models.Field.to_python

        :raises django.core.exceptions.ValidationError:
            if the data can't be converted

        """
        if value is None or isinstance(value, Rut):
            converted_value = value
        else:
            try:
                converted_value = Rut(value, validate_dv=False)  # type: ignore
            except (AttributeError, TypeError, ValueError):
                raise django.core.exceptions.ValidationError(
                    self.error_messages['invalid'],
                    code='invalid',
                    params={'value': value},
                )

        return converted_value

    def value_to_string(self, obj: django.db.models.Model) -> str:
        """
        Convert to a string the field value of model instance``obj``.

        Used to serialize the value of the field.

        .. seealso::
            https://docs.djangoproject.com/en/2.1/howto/custom-model-fields/#converting-field-data-for-serialization

        """
        # note: according to official docs, 'value_from_object' is the
        #   "best way to get the field's value prior to serialization".
        value: Optional[Rut] = self.value_from_object(obj)

        return '' if value is None else value.canonical



"""
Helpers for parsing data from a "cesión"'s AEC XML doc.


Usage:

>>> from cl_sii.libs import xml_utils

>>> aec_xml_file_path = '/dir/my_aec.xml'
>>> with open(aec_xml_file_path, mode='rb') as f:
...     xml_doc = xml_utils.parse_untrusted_xml(f.read())

>>> validate_aec_xml(xml_doc)
>>> aec_xml = parse_aec_xml(xml_doc)
"""

from __future__ import annotations

import logging
from datetime import date, datetime
from pathlib import Path
from typing import Mapping, Optional, Sequence

import pydantic

import cl_sii.dte.data_models
import cl_sii.dte.parse
from cl_sii.dte.constants import TipoDte
from cl_sii.dte.data_models import DteXmlData, is_input_trusted_according_to_validation_context
from cl_sii.dte.parse import DTE_XMLNS_MAP
from cl_sii.libs import encoding_utils, tz_utils, xml_utils
from cl_sii.libs.xml_utils import XmlElement
from cl_sii.rut import Rut
from . import data_models_aec


logger = logging.getLogger(__name__)


_AEC_XML_SCHEMA_PATH = Path(
    Path(__file__).parent.parent,
    Path('data/ref/factura_electronica/schemas-xml/AEC_v10.xsd'),
).resolve()

AEC_XML_SCHEMA_OBJ = xml_utils.read_xml_schema(str(_AEC_XML_SCHEMA_PATH))
"""
XML schema obj for AEC XML document validation.

..seealso::
    XML schema of ``{http://www.sii.cl/SiiDte}/AEC`` in
    'data/ref/factura_electronica/schemas-xml/AEC_v10.xsd' (c7adc5a2)

It is read from a file at import time to avoid unnecessary reads afterwards.
"""


###############################################################################
# Main Functions
###############################################################################


def validate_aec_xml(xml_doc: XmlElement) -> None:
    """
    Validate ``xml_doc`` against AEC's XML schema.

    :raises xml_utils.XmlSchemaDocValidationError:
    """
    # TODO: Add better and more precise exception handling.
    xml_utils.validate_xml_doc(AEC_XML_SCHEMA_OBJ, xml_doc)


def parse_aec_xml(xml_doc: XmlElement, trust_input: bool = False) -> data_models_aec.AecXml:
    """
    Parse data from a "cesión"'s AEC XML doc.

    .. warning::
        It is assumed that ``xml_doc`` is an ``{http://www.sii.cl/SiiDte}/AEC`` XML element.

    :param xml_doc:
        AEC XML document.
    :param trust_input:
        If ``True``, the input data is trusted to be valid and
        some validation errors are replaced by warnings.

        .. warning::
            Use this option *only* if you obtained the AEC XML document
            from the SII *and* you need to work around some validation errors
            that the SII should have caught, but let through.
    """
    aec_struct = _Aec.parse_xml(xml_doc, trust_input=trust_input)
    return aec_struct.as_aec_xml(trust_input=trust_input)


###############################################################################
# Parser Functions and Models
###############################################################################


def _empty_str_to_none(v: object) -> object:
    """
    Reusable Pydantic validator that converts empty strings to ``None``.
    """
    if isinstance(v, str):
        if not v.strip():
            v = None
    return v


def _validate_rut(v: object) -> object:
    """
    Reusable Pydantic validator for fields of type :class:`Rut`.
    """
    if isinstance(v, str):
        v = Rut(value=v, validate_dv=False)  # Raises ValueError if invalid.
    return v


class _XmlSignature(pydantic.BaseModel):
    """
    Parser for ``//Signature``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    signature_value: bytes
    key_info_x509_data_x509_cert: bytes

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: //Signature/KeyInfo
        key_info_em = xml_em.find('ds:KeyInfo', namespaces=xml_utils.XML_DSIG_NS_MAP)
        assert key_info_em is not None

        # XPath: //Signature/KeyInfo/X509Data
        key_info_x509_data_em = key_info_em.find(
            'ds:X509Data', namespaces=xml_utils.XML_DSIG_NS_MAP
        )
        assert key_info_x509_data_em is not None

        # XPath: //Signature
        return dict(
            # XPath: //Signature/SignatureValue
            signature_value=xml_em.findtext(
                'ds:SignatureValue',
                namespaces=xml_utils.XML_DSIG_NS_MAP,
            ),
            #
            # XPath: //Signature/KeyInfo/X509Data/X509Certificate
            key_info_x509_data_x509_cert=key_info_x509_data_em.findtext(
                'ds:X509Certificate',
                namespaces=xml_utils.XML_DSIG_NS_MAP,
            ),
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator(
        'signature_value',
        'key_info_x509_data_x509_cert',
        mode='before',
    )
    @classmethod
    def validate_base64(cls, v: object) -> object:
        if isinstance(v, (str, bytes)):
            v = encoding_utils.decode_base64_strict(v)  # Raises ValueError.
        return v

    # Note: Even though this validation seems to make perfect sense, there are some
    # real cases of SII-approved AEC where this is not fulfilled.
    # We will keep this validation in case we need it in the future.
    # @pydantic.validator('key_info_x509_data_x509_cert')
    # def validate_certificate_is_loadable(cls, v: object) -> object:
    #     if isinstance(v, bytes):
    #         _ = crypto_utils.load_der_x509_cert(v)  # Raises ValueError.
    #     return v


class _Cesionario(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cesionario``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    rut: Rut
    razon_social: str
    direccion: str
    email: str

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cesionario
        return dict(
            rut=xml_em.findtext('sii-dte:RUT', namespaces=DTE_XMLNS_MAP),
            razon_social=xml_em.findtext('sii-dte:RazonSocial', namespaces=DTE_XMLNS_MAP),
            direccion=xml_em.findtext('sii-dte:Direccion', namespaces=DTE_XMLNS_MAP),
            email=xml_em.findtext('sii-dte:eMail', namespaces=DTE_XMLNS_MAP),
        )

    ###########################################################################
    # Validators
    ###########################################################################

    _validate_rut = pydantic.field_validator(
        'rut',
        mode='before',
    )(_validate_rut)


class _RutAutorizado(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente/RUTAutorizado``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    rut: Rut
    nombre: Optional[str]

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente/RUTAutorizado
        return dict(
            rut=xml_em.findtext('sii-dte:RUT', namespaces=DTE_XMLNS_MAP),
            nombre=xml_em.findtext('sii-dte:Nombre', namespaces=DTE_XMLNS_MAP),
        )

    ###########################################################################
    # Validators
    ###########################################################################

    _empty_str_to_none = pydantic.field_validator(
        'nombre',
        mode='before',
    )(_empty_str_to_none)

    _validate_rut = pydantic.field_validator(
        'rut',
        mode='before',
    )(_validate_rut)


class _Cedente(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    rut: Rut
    razon_social: str
    direccion: str
    email: str
    ruts_autorizados: Sequence[_RutAutorizado]  # 1..3 occurrences
    declaracion_jurada: Optional[str]

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente/RUTAutorizado
        cedente_personas_autorizadas_em = xml_em.findall(
            'sii-dte:RUTAutorizado',
            namespaces=DTE_XMLNS_MAP,
        )
        cedente_persona_autorizada_dict_list = [
            _RutAutorizado.parse_xml_to_dict(cedente_persona_autorizada)
            for cedente_persona_autorizada in cedente_personas_autorizadas_em
        ]

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente
        return dict(
            rut=xml_em.findtext('sii-dte:RUT', namespaces=DTE_XMLNS_MAP),
            razon_social=xml_em.findtext('sii-dte:RazonSocial', namespaces=DTE_XMLNS_MAP),
            direccion=xml_em.findtext('sii-dte:Direccion', namespaces=DTE_XMLNS_MAP),
            email=xml_em.findtext('sii-dte:eMail', namespaces=DTE_XMLNS_MAP),
            declaracion_jurada=xml_em.findtext(
                'sii-dte:DeclaracionJurada',
                namespaces=DTE_XMLNS_MAP,
            )
            or None,
            ruts_autorizados=cedente_persona_autorizada_dict_list,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('rut', mode='before')
    @classmethod
    def validate_rut(cls, v: object) -> object:
        if isinstance(v, str):
            v = Rut(value=v, validate_dv=False)  # Raises ValueError if invalid.
        return v

    @pydantic.field_validator('ruts_autorizados')
    @classmethod
    def validate_ruts_autorizados_item_count(cls, v: Sequence) -> object:
        if len(v) < 1:
            raise ValueError("must contain at least one item")
        if len(v) > 3:
            raise ValueError("must contain at most three items")
        return v


class _IdDte(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/IdDTE``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    rut_emisor: Rut
    tipo_dte: TipoDte
    folio: int
    fch_emis: date
    rut_receptor: Rut
    mnt_total: int

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/IdDTE
        return dict(
            rut_emisor=xml_em.findtext('sii-dte:RUTEmisor', namespaces=DTE_XMLNS_MAP),
            tipo_dte=xml_em.findtext('sii-dte:TipoDTE', namespaces=DTE_XMLNS_MAP),
            folio=xml_em.findtext('sii-dte:Folio', namespaces=DTE_XMLNS_MAP),
            fch_emis=xml_em.findtext('sii-dte:FchEmis', namespaces=DTE_XMLNS_MAP),
            rut_receptor=xml_em.findtext('sii-dte:RUTReceptor', namespaces=DTE_XMLNS_MAP),
            mnt_total=xml_em.findtext('sii-dte:MntTotal', namespaces=DTE_XMLNS_MAP),
        )

    def as_dte_data_l1(self) -> cl_sii.dte.data_models.DteDataL1:
        return cl_sii.dte.data_models.DteDataL1(
            emisor_rut=self.rut_emisor,
            tipo_dte=self.tipo_dte,
            folio=self.folio,
            fecha_emision_date=self.fch_emis,
            receptor_rut=self.rut_receptor,
            monto_total=self.mnt_total,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    _validate_rut_emisor = pydantic.field_validator(
        'rut_emisor',
        mode='before',
    )(_validate_rut)

    _validate_rut_receptor = pydantic.field_validator(
        'rut_receptor',
        mode='before',
    )(_validate_rut)

    @pydantic.field_validator('tipo_dte', mode='before')
    @classmethod
    def validate_tipo_dte(cls, v: object) -> object:
        if isinstance(v, str):
            v = int(v)  # Raises ValueError if invalid.
        if isinstance(v, int):
            v = TipoDte(v)  # Raises ValueError if invalid.
        return v


class _DocumentoCesion(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    # id: str
    """
    This value seems to be worthless (only useful for internal references in the XML doc).
    e.g. 'CES7393a78afa6c4f709ee9e9e943cb50a3', 'HEF_CESION_T33F170_SEQ2'
    """

    seq_cesion: int
    id_dte: _IdDte
    cedente: _Cedente
    cesionario: _Cesionario
    monto_cesion: int
    ultimo_vencimiento: date
    tmst_cesion: datetime
    email_deudor: Optional[str]

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/IdDTE
        id_dte_em = xml_em.find('sii-dte:IdDTE', namespaces=DTE_XMLNS_MAP)
        assert id_dte_em is not None
        id_dte_dict = _IdDte.parse_xml_to_dict(id_dte_em)

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cedente
        cedente_em = xml_em.find('sii-dte:Cedente', namespaces=DTE_XMLNS_MAP)
        assert cedente_em is not None
        cedente_dict = _Cedente.parse_xml_to_dict(cedente_em)

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion/Cesionario
        cesionario_em = xml_em.find('sii-dte:Cesionario', namespaces=DTE_XMLNS_MAP)
        assert cesionario_em is not None
        cesionario_dict = _Cesionario.parse_xml_to_dict(cesionario_em)

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion
        return dict(
            # id=xml_em.get('ID'),
            seq_cesion=xml_em.findtext('sii-dte:SeqCesion', namespaces=DTE_XMLNS_MAP),
            id_dte=id_dte_dict,
            cedente=cedente_dict,
            cesionario=cesionario_dict,
            monto_cesion=xml_em.findtext('sii-dte:MontoCesion', namespaces=DTE_XMLNS_MAP),
            ultimo_vencimiento=xml_em.findtext(
                'sii-dte:UltimoVencimiento',
                namespaces=DTE_XMLNS_MAP,
            ),
            tmst_cesion=xml_em.findtext('sii-dte:TmstCesion', namespaces=DTE_XMLNS_MAP),
            email_deudor=xml_em.findtext('sii-dte:eMailDeudor', namespaces=DTE_XMLNS_MAP) or None,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('tmst_cesion')
    @classmethod
    def validate_datetime(cls, v: datetime) -> datetime:
        if isinstance(v, datetime):
            v = tz_utils.convert_naive_dt_to_tz_aware(
                dt=v,
                tz=data_models_aec.CesionAecXml.DATETIME_FIELDS_TZ,
            )
        return v


class _Cesion(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/Cesion``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    documento_cesion: _DocumentoCesion
    # signature: _XmlSignature

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/DocumentoCesion
        doc_cesion_em = xml_em.find('sii-dte:DocumentoCesion', namespaces=DTE_XMLNS_MAP)
        assert doc_cesion_em is not None
        doc_cesion_dict = _DocumentoCesion.parse_xml_to_dict(doc_cesion_em)

        # Signature over 'DocumentoCesion'
        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion/Signature
        # signature_over_doc_cesion_em = xml_em.find(
        #     'ds:Signature',
        #     namespaces=xml_utils.XML_DSIG_NS_MAP,
        # )
        # signature_over_doc_cesion_dict = _XmlSignature.parse_xml_to_dict(
        #     signature_over_doc_cesion_em,
        # )

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion
        return dict(
            documento_cesion=doc_cesion_dict,
            # signature=signature_over_doc_cesion_dict,
        )

    def as_cesion_aec_xml(self) -> data_models_aec.CesionAecXml:
        doc_cesion_struct = self.documento_cesion
        # signature_over_doc_cesion_struct = self.signature  # noqa: F841
        cesion_dte_struct = doc_cesion_struct.id_dte.as_dte_data_l1()
        cedente_persona_autorizada_struct_first = doc_cesion_struct.cedente.ruts_autorizados[0]

        return data_models_aec.CesionAecXml(
            dte=cesion_dte_struct,
            seq=doc_cesion_struct.seq_cesion,
            cedente_rut=doc_cesion_struct.cedente.rut,
            cesionario_rut=doc_cesion_struct.cesionario.rut,
            monto_cesion=doc_cesion_struct.monto_cesion,
            fecha_cesion_dt=doc_cesion_struct.tmst_cesion,
            fecha_ultimo_vencimiento=doc_cesion_struct.ultimo_vencimiento,
            cedente_razon_social=doc_cesion_struct.cedente.razon_social,
            cedente_direccion=doc_cesion_struct.cedente.direccion,
            cedente_email=doc_cesion_struct.cedente.email,
            cedente_persona_autorizada_rut=cedente_persona_autorizada_struct_first.rut,
            cedente_persona_autorizada_nombre=cedente_persona_autorizada_struct_first.nombre,
            cesionario_razon_social=doc_cesion_struct.cesionario.razon_social,
            cesionario_direccion=doc_cesion_struct.cesionario.direccion,
            cesionario_email=doc_cesion_struct.cesionario.email,
            dte_deudor_email=doc_cesion_struct.email_deudor,
            cedente_declaracion_jurada=doc_cesion_struct.cedente.declaracion_jurada,
            # signature_value=signature_over_doc_cesion_struct.signature_value,
            # signature_x509_cert_der=signature_over_doc_cesion_struct.key_info_x509_data_x509_cert,
        )


class _DocumentoDteCedido(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/DTECedido/DocumentoDTECedido``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    # id: str
    """
    This value seems to be worthless (only useful for internal references in the XML doc).
      e.g. 'DTE6e0d95997db9489aa4cdd768a45852f6', 'DTE5716484782d745b6822257822a3536d1'
    """

    dte: DteXmlData
    # tmst_firma: datetime

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido/DocumentoDTECedido/DTE
        dte_em = xml_em.find('sii-dte:DTE', namespaces=DTE_XMLNS_MAP)

        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido/DocumentoDTECedido
        return dict(
            # id=xml_em.get('ID'),
            dte=dte_em,
            # tmst_firma=xml_em.findtext('sii-dte:TmstFirma', namespaces=DTE_XMLNS_MAP),
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('dte', mode='before')
    @classmethod
    def validate_dte(cls, v: object, info: pydantic.ValidationInfo) -> object:
        if isinstance(v, XmlElement):
            cl_sii.dte.parse.validate_dte_xml(v)
            trust_dte = is_input_trusted_according_to_validation_context(info.context)
            v = cl_sii.dte.parse.parse_dte_xml(v, trust_input=trust_dte)
        return v

    # @pydantic.validator('tmst_firma')
    # def validate_datetime(cls, v: object) -> object:
    #     if isinstance(v, str):
    #         v = datetime.fromisoformat(v)
    #
    #     if isinstance(v, datetime):
    #         v = tz_utils.convert_naive_dt_to_tz_aware(
    #             dt=v,
    #             tz=data_models_aec.AecXml.DATETIME_FIELDS_TZ,
    #         )
    #     return v


class _DteCedido(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Cesiones/DTECedido``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    documento_dte_cedido: _DocumentoDteCedido
    # signature: _XmlSignature

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido/DocumentoDTECedido
        doc_dte_cedido_em = xml_em.find(
            'sii-dte:DocumentoDTECedido',
            namespaces=DTE_XMLNS_MAP,
        )
        assert doc_dte_cedido_em is not None

        # Signature over 'DocumentoDTECedido'
        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido/Signature
        # signature_over_doc_dte_cedido_em = xml_em.find(
        #     'ds:Signature',
        #     namespaces=xml_utils.XML_DSIG_NS_MAP,
        # )
        # signature_over_doc_dte_cedido_dict = _XmlSignature.parse_xml_to_dict(
        #     signature_over_doc_dte_cedido_em,
        # )

        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido/DocumentoDTECedido
        doc_dte_cedido_dict = _DocumentoDteCedido.parse_xml_to_dict(doc_dte_cedido_em)

        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido
        return dict(
            documento_dte_cedido=doc_dte_cedido_dict,
            # signature=signature_over_doc_dte_cedido_dict,
        )


class _Caratula(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC/Caratula``.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra='forbid',
        frozen=True,
        str_min_length=1,
        str_strip_whitespace=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    rut_cedente: Rut
    rut_cesionario: Rut
    nmb_contacto: Optional[str]
    fono_contacto: Optional[str]
    mail_contacto: Optional[str]
    tmst_firmaenvio: datetime

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Caratula
        return dict(
            rut_cedente=xml_em.findtext('sii-dte:RutCedente', namespaces=DTE_XMLNS_MAP),
            rut_cesionario=xml_em.findtext('sii-dte:RutCesionario', namespaces=DTE_XMLNS_MAP),
            nmb_contacto=xml_em.findtext('sii-dte:NmbContacto', namespaces=DTE_XMLNS_MAP) or None,
            fono_contacto=xml_em.findtext('sii-dte:FonoContacto', namespaces=DTE_XMLNS_MAP) or None,
            mail_contacto=xml_em.findtext('sii-dte:MailContacto', namespaces=DTE_XMLNS_MAP) or None,
            tmst_firmaenvio=xml_em.findtext('sii-dte:TmstFirmaEnvio', namespaces=DTE_XMLNS_MAP),
        )

    ###########################################################################
    # Validators
    ###########################################################################

    _empty_str_to_none = pydantic.field_validator(
        'nmb_contacto',
        'fono_contacto',
        'mail_contacto',
        mode='before',
    )(_empty_str_to_none)

    _validate_rut_cedente = pydantic.field_validator(
        'rut_cedente',
        mode='before',
    )(_validate_rut)

    _validate_rut_cesionario = pydantic.field_validator(
        'rut_cesionario',
        mode='before',
    )(_validate_rut)

    @pydantic.field_validator('tmst_firmaenvio')
    @classmethod
    def validate_datetime(cls, v: datetime) -> datetime:
        v = tz_utils.convert_naive_dt_to_tz_aware(
            dt=v,
            tz=data_models_aec.AecXml.DATETIME_FIELDS_TZ,
        )
        return v


class _DocumentoAec(pydantic.BaseModel):
    """
    Parser for ``/AEC/DocumentoAEC``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    # id: str
    """
    This value seems to be worthless (only useful for internal references in the XML doc).
      e.g. 'HEF_AEC_T33F170_SEQ2', 'AEC1589423e81824cdcbfd2f0f4496f2dfb'
    """

    caratula: _Caratula
    cesiones_dte_cedido: _DteCedido
    cesiones_cesion: Sequence[_Cesion]

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @staticmethod
    def parse_xml_to_dict(xml_em: XmlElement) -> Mapping[str, object]:
        """
        Parse XML element and return a dictionary.
        """
        # XPath: /AEC/DocumentoAEC/Caratula
        caratula_em = xml_em.find('sii-dte:Caratula', namespaces=DTE_XMLNS_MAP)
        assert caratula_em is not None
        caratula_dict = _Caratula.parse_xml_to_dict(caratula_em)

        # XPath: /AEC/DocumentoAEC/Cesiones
        cesiones_em = xml_em.find('sii-dte:Cesiones', namespaces=DTE_XMLNS_MAP)
        assert cesiones_em is not None

        # XPath: /AEC/DocumentoAEC/Cesiones/DTECedido
        dte_cedido_em = cesiones_em.find('sii-dte:DTECedido', namespaces=DTE_XMLNS_MAP)
        assert dte_cedido_em is not None
        dte_cedido_dict = _DteCedido.parse_xml_to_dict(dte_cedido_em)

        # XPath: /AEC/DocumentoAEC/Cesiones/Cesion
        cesion_em_list: Sequence[XmlElement] = cesiones_em.findall(
            'sii-dte:Cesion',
            namespaces=DTE_XMLNS_MAP,
        )
        cesion_dict_list: Sequence[Mapping[str, object]]
        cesion_dict_list = [_Cesion.parse_xml_to_dict(cesion_em) for cesion_em in cesion_em_list]

        # XPath: /AEC/DocumentoAEC
        return dict(
            # id=xml_em.get('ID'),
            caratula=caratula_dict,
            cesiones_dte_cedido=dte_cedido_dict,
            cesiones_cesion=cesion_dict_list,
        )

    ###########################################################################
    # Validators
    ###########################################################################

    @pydantic.field_validator('cesiones_cesion')
    @classmethod
    def validate_cesiones_cesion_min_items(cls, v: object) -> object:
        if isinstance(v, Sequence):
            if len(v) < 1:
                raise ValueError("must contain at least one item")
        return v


class _Aec(pydantic.BaseModel):
    """
    Parser for ``/AEC``.
    """

    model_config = pydantic.ConfigDict(
        extra='forbid',
        frozen=True,
    )

    ###########################################################################
    # Fields
    ###########################################################################

    documento_aec: _DocumentoAec
    signature: _XmlSignature

    ###########################################################################
    # Custom Methods
    ###########################################################################

    @classmethod
    def parse_xml(cls, xml_doc: XmlElement, trust_input: bool = False) -> _Aec:
        aec_dict = cls.parse_xml_to_dict(xml_doc)
        return cls.model_validate(
            aec_dict,
            context={
                cl_sii.dte.data_models.VALIDATION_CONTEXT_TRUST_INPUT: trust_input,
            },
        )

    def as_aec_xml(self, trust_input: bool = False) -> data_models_aec.AecXml:
        doc_aec_struct = self.documento_aec
        signature_over_doc_aec_struct = self.signature

        caratula_struct = doc_aec_struct.caratula
        dte = doc_aec_struct.cesiones_dte_cedido.documento_dte_cedido.dte
        cesion_struct_list = doc_aec_struct.cesiones_cesion

        aec_xml_cesion_list: Sequence[data_models_aec.CesionAecXml]
        aec_xml_cesion_list = [
            cesion_struct.as_cesion_aec_xml() for cesion_struct in cesion_struct_list
        ]

        return data_models_aec.AEC_XML_PYDANTIC_TYPE_ADAPTER.validate_python(
            dict(
                dte=dte,
                cedente_rut=caratula_struct.rut_cedente,
                cesionario_rut=caratula_struct.rut_cesionario,
                fecha_firma_dt=caratula_struct.tmst_firmaenvio,
                signature_value=signature_over_doc_aec_struct.signature_value,
                signature_x509_cert_der=signature_over_doc_aec_struct.key_info_x509_data_x509_cert,
                cesiones=aec_xml_cesion_list,
                contacto_nombre=caratula_struct.nmb_contacto,
                contacto_telefono=caratula_struct.fono_contacto,
                contacto_email=caratula_struct.mail_contacto,
            ),
            context={
                cl_sii.dte.data_models.VALIDATION_CONTEXT_TRUST_INPUT: trust_input,
            },
        )

    @staticmethod
    def parse_xml_to_dict(xml_doc: XmlElement) -> Mapping[str, object]:
        """
        Parse data from a "cesión"'s AEC XML doc and return a dictionary.
        """
        # XPath: /AEC
        aec_em = xml_doc

        # XPath: /AEC/DocumentoAEC
        doc_aec_em = aec_em.find('sii-dte:DocumentoAEC', namespaces=DTE_XMLNS_MAP)
        assert doc_aec_em is not None
        doc_aec_dict = _DocumentoAec.parse_xml_to_dict(doc_aec_em)

        # Signature over 'DocumentoAEC'
        # XPath: /AEC/Signature
        signature_over_doc_aec_em = aec_em.find(
            'ds:Signature',
            namespaces=xml_utils.XML_DSIG_NS_MAP,
        )
        assert signature_over_doc_aec_em is not None
        signature_over_doc_aec_dict = _XmlSignature.parse_xml_to_dict(signature_over_doc_aec_em)

        # XPath: /AEC
        return dict(
            documento_aec=doc_aec_dict,
            signature=signature_over_doc_aec_dict,
        )



"""
cl_sii "extras" / Pydantic types.
"""

from __future__ import annotations

import re
import sys
from typing import Any, ClassVar, Pattern


if sys.version_info[:2] >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

try:
    import pydantic
    import pydantic.json_schema
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'pydantic' is required to use this module.") from exc

try:
    import pydantic_core
except ImportError as exc:  # pragma: no cover
    raise ImportError("Package 'pydantic-core' is required to use this module.") from exc

import cl_sii.rut
import cl_sii.rut.constants


class _RutPydanticAnnotation:
    """
    `Annotated` wrapper that can be used as the annotation for `cl_sii.rut.Rut`
    fields on `pydantic.BaseModels`, `@pydantic.dataclasses`, etc.

    .. seealso::
        - Handling third-party types:
            https://docs.pydantic.dev/2.9/concepts/types/#handling-third-party-types
            (https://github.com/pydantic/pydantic/blob/v2.9.2/docs/concepts/types.md#handling-third-party-types)
        - Customizing the core schema and JSON schema:
            https://docs.pydantic.dev/2.9/architecture/#customizing-the-core-schema-and-json-schema
            (https://github.com/pydantic/pydantic/blob/v2.9.2/docs/architecture.md#customizing-the-core-schema-and-json-schema)
        - Implementing __get_pydantic_json_schema__:
            https://docs.pydantic.dev/2.9/concepts/json_schema/#implementing-__get_pydantic_json_schema__
            (https://github.com/pydantic/pydantic/blob/v2.9.2/docs/concepts/json_schema.md#implementing-__get_pydantic_json_schema__-)

    Examples:

    >>> from typing import Annotated
    >>> import pydantic
    >>> import cl_sii.rut

    >>> Rut = Annotated[cl_sii.rut.Rut, _RutPydanticAnnotation]

    >>> class ExampleModel(pydantic.BaseModel):
    ...     rut: Rut
    >>>
    >>> example_model_instance = ExampleModel.model_validate({'rut': '78773510-K'})

    >>> import pydantic.dataclasses
    >>>
    >>> @pydantic.dataclasses.dataclass
    ... class ExampleDataclass:
    ...     rut: Rut
    >>>
    >>> example_dataclass_instance = ExampleDataclass(rut='78773510-K')

    >>> example_type_adapter = pydantic.TypeAdapter(Rut)
    >>>
    >>> example_type_adapter.validate_python('78773510-K')
    Rut('78773510-K')
    >>> example_type_adapter.validate_json('"78773510-K"')
    Rut('78773510-K')
    >>> example_type_adapter.dump_python(cl_sii.rut.Rut('78773510-K'))
    '78773510-K'
    >>> example_type_adapter.dump_json(cl_sii.rut.Rut('78773510-K'))
    b'"78773510-K"'
    >>> example_json_schema = example_type_adapter.json_schema()
    """

    RUT_CANONICAL_STRICT_REGEX: ClassVar[Pattern] = re.compile(
        re.sub(
            pattern=r'\?P<\w+>',
            repl='',
            string=cl_sii.rut.constants.RUT_CANONICAL_STRICT_REGEX.pattern,
        )
    )
    """
    RUT (strict) regex for canonical format, without named groups.

    .. warning::
        JSON Schema and OpenAPI use the regular expression syntax from
        JavaScript (ECMA 262), which does not support Python’s named groups.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        def validate_from_str(value: str) -> cl_sii.rut.Rut:
            return cl_sii.rut.Rut(value, validate_dv=False)

        from_str_schema = pydantic_core.core_schema.chain_schema(
            [
                cls.str_schema(),
                pydantic_core.core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return pydantic_core.core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=pydantic_core.core_schema.union_schema(
                [
                    pydantic_core.core_schema.is_instance_schema(cl_sii.rut.Rut),
                    from_str_schema,
                ]
            ),
            serialization=pydantic_core.core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.canonical,
                when_used='unless-none',
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: pydantic_core.core_schema.CoreSchema,
        handler: pydantic.GetJsonSchemaHandler,
    ) -> pydantic.json_schema.JsonSchemaValue:
        core_schema = cls.str_schema()
        return handler(core_schema)

    @classmethod
    def str_schema(cls) -> pydantic_core.core_schema.CoreSchema:
        return pydantic_core.core_schema.str_schema(pattern=cls.RUT_CANONICAL_STRICT_REGEX)


Rut = Annotated[cl_sii.rut.Rut, _RutPydanticAnnotation]
"""
Convenience type alias for Pydantic fields that represent Chilean RUTs.
"""
